# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Convenience wrapper for GP prediction from indices.
"""

import numpy as np

from typing import Callable, Optional, Tuple, Union

from MuyGPyS.gp.tensors import (
    crosswise_tensor,
    make_predict_tensors,
    make_train_tensors,
)
from MuyGPyS.gp import MuyGPS, MultivariateMuyGPS as MMuyGPS
from MuyGPyS.optimize import optimize_from_tensors
from MuyGPyS.optimize.loss import lool_fn


def tensors_from_indices(
    muygps: Union[MuyGPS, MMuyGPS],
    indices: np.ndarray,
    nn_indices: np.ndarray,
    test: np.ndarray,
    train: np.ndarray,
    targets: np.ndarray,
    **kwargs,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    crosswise_tensor, pairwise_tensor, batch_nn_targets = make_predict_tensors(
        indices, nn_indices, test, train, targets
    )
    if isinstance(muygps, MuyGPS):
        pairwise_tensor = muygps.kernel(pairwise_tensor)
        crosswise_tensor = muygps.kernel(crosswise_tensor)
    return pairwise_tensor, crosswise_tensor, batch_nn_targets


def posterior_mean_from_indices(
    muygps: Union[MuyGPS, MMuyGPS],
    indices: np.ndarray,
    nn_indices: np.ndarray,
    test: np.ndarray,
    train: np.ndarray,
    targets: np.ndarray,
    **kwargs,
) -> np.ndarray:
    pairwise_tensor, crosswise_tensor, batch_nn_targets = tensors_from_indices(
        muygps, indices, nn_indices, test, train, targets
    )
    return muygps.posterior_mean(
        pairwise_tensor, crosswise_tensor, batch_nn_targets
    )


def posterior_variance_from_indices(
    muygps: Union[MuyGPS, MMuyGPS],
    indices: np.ndarray,
    nn_indices: np.ndarray,
    test: np.ndarray,
    train: np.ndarray,
    targets: np.ndarray,
    **kwargs,
) -> np.ndarray:
    pairwise_tensor, crosswise_tensor, batch_nn_targets = tensors_from_indices(
        muygps, indices, nn_indices, test, train, targets
    )
    return muygps.posterior_variance(
        pairwise_tensor, crosswise_tensor, **kwargs
    )


def regress_from_indices(
    muygps: Union[MuyGPS, MMuyGPS],
    indices: np.ndarray,
    nn_indices: np.ndarray,
    test: np.ndarray,
    train: np.ndarray,
    targets: np.ndarray,
    **kwargs,
) -> Tuple[np.ndarray, np.ndarray]:
    pairwise_tensor, crosswise_tensor, batch_nn_targets = tensors_from_indices(
        muygps, indices, nn_indices, test, train, targets
    )
    return muygps.posterior_mean(
        pairwise_tensor, crosswise_tensor, batch_nn_targets
    ), muygps.posterior_variance(pairwise_tensor, crosswise_tensor, **kwargs)


def fast_posterior_mean_from_indices(
    muygps: Union[MuyGPS, MMuyGPS],
    indices: np.ndarray,
    nn_indices: np.ndarray,
    test_features: np.ndarray,
    train_features: np.ndarray,
    closest_index: np.ndarray,
    coeffs_tensor: np.ndarray,
) -> np.ndarray:
    crosswise_diffs = crosswise_tensor(
        test_features,
        train_features,
        indices,
        nn_indices,
    )
    if isinstance(muygps, MuyGPS):
        Kcross = muygps.kernel(crosswise_diffs)
        return muygps.fast_posterior_mean(
            Kcross,
            coeffs_tensor[closest_index, :, :],
        )
    else:
        return muygps.fast_posterior_mean(
            crosswise_diffs,
            coeffs_tensor[closest_index, :, :],
        )


def optimize_from_indices(
    muygps: MuyGPS,
    batch_indices: np.ndarray,
    batch_nn_indices: np.ndarray,
    train_features: np.ndarray,
    train_targets: np.ndarray,
    loss_fn: Callable = lool_fn,
    obj_method: str = "loo_crossval",
    opt_method: str = "bayes",
    sigma_method: Optional[str] = "analytic",
    verbose: bool = False,
    **kwargs,
) -> MuyGPS:
    """
    Find an optimal model directly from the data.

    Use this method if you do not need to retain the difference and kernel
    tensors used for optimization.

    See the following example, where we have already created a `batch_indices`
    vector and a `batch_nn_indices` matrix using
    :class:`MuyGPyS.neighbors.NN_Wrapper`, and initialized a
    :class:`~MuyGPyS.gp.muygps.MuyGPS` model `muygps`.

    Example:
        >>> from MuyGPyS.optimize.chassis import optimize_from_indices
        >>> muygps = optimize_from_indices(
        ...         muygps,
        ...         batch_indices,
        ...         batch_nn_indices,
        ...         train_features,
        ...         train_features,
        ...         train_responses,
        ...         loss_fn=lool_fn,
        ...         obj_method='loo_crossval',
        ...         opt_method='scipy',
        ...         verbose=True,
        ... )
        parameters to be optimized: ['nu']
        bounds: [[0.1 1. ]]
        sampled x0: [0.8858425]
        optimizer results:
              fun: 0.4797763813693626
         hess_inv: <1x1 LbfgsInvHessProduct with dtype=float64>
              jac: array([-3.06976666e-06])
          message: b'CONVERGENCE: NORM_OF_PROJECTED_GRADIENT_<=_PGTOL'
             nfev: 16
              nit: 5
             njev: 8
           status: 0
          success: True
                x: array([0.39963594])

    Args:
        muygps:
            The model to be optimized.
        batch_indices:
            A vector of integers of shape `(batch_count,)` identifying the
            training batch of observations to be approximated.
        batch_nn_indices:
            A matrix of integers of shape `(batch_count, nn_count)` listing the
            nearest neighbor indices for all observations in the batch.
        train_features:
            The full floating point training data matrix of shape
            `(train_count, feature_count)`.
        train_targets:
            A matrix of shape `(train_count, feature_count)` whose rows are
            vector-valued responses for each training element.
        loss_fn:
            Indicates the loss functor to be used.
        obj_method:
            Indicates the objective function to be minimized. Currently
            restricted to `"loo_crossval"`.
        opt_method:
            Indicates the optimization method to be used. Currently restricted
            to `"bayesian"` (alternately `"bayes"` or `"bayes_opt"`) and
            `"scipy"`.
        sigma_method:
            The optimization method to be employed to learn the `sigma_sq`
            hyperparameter.
        verbose:
            If True, print debug messages.
        kwargs:
            Additional keyword arguments to be passed to the wrapper optimizer.

    Returns:
        A new MuyGPs model whose specified hyperparameters have been optimized.
    """
    (
        crosswise_diffs,
        pairwise_diffs,
        batch_targets,
        batch_nn_targets,
    ) = make_train_tensors(
        batch_indices,
        batch_nn_indices,
        train_features,
        train_targets,
    )
    return optimize_from_tensors(
        muygps,
        batch_targets,
        batch_nn_targets,
        crosswise_diffs,
        pairwise_diffs,
        loss_fn=loss_fn,
        obj_method=obj_method,
        opt_method=opt_method,
        sigma_method=sigma_method,
        verbose=verbose,
        **kwargs,
    )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from typing import Callable, Dict, Generator, Optional, Tuple, Type, Union

import MuyGPyS._src.math as mm
import MuyGPyS._src.math.numpy as np
from MuyGPyS import config
from MuyGPyS._src.mpi_utils import _is_mpi_mode

if config.muygpys_hnswlib_enabled is True:  # type: ignore
    _basic_nn_kwarg_options = [
        {"nn_method": "exact", "algorithm": "ball_tree"},
        {
            "nn_method": "hnsw",
            "space": "l2",
            "ef_construction": 100,
            "M": 16,
        },
    ]
else:
    _basic_nn_kwarg_options = [
        {"nn_method": "exact", "algorithm": "ball_tree"},
    ]

_exact_nn_kwarg_options = ({"nn_method": "exact", "algorithm": "ball_tree"},)

_basic_opt_method_and_kwarg_options = [
    ["scipy", dict()],
    [
        "bayesian",
        {
            "random_state": 1,
            "init_points": 3,
            "n_iter": 10,
            "allow_duplicate_points": True,
        },
    ],
]

_advanced_opt_method_and_kwarg_options = [
    ["scipy", dict()],
    [
        "bayesian",
        {
            "random_state": 1,
            "init_points": 5,
            "n_iter": 20,
            "allow_duplicate_points": True,
        },
    ],
]


def _sq_rel_err(
    tru: Union[float, mm.ndarray], est: Union[float, mm.ndarray]
) -> float:
    """
    Compute the relative squared error between two arguments.

    Args:
        tru:
            The approximated quantity.
        est:
            The estimate to be evaluated.

    Returns:
        An i.i.d. Gaussian matrix of shape `(data_count, feature_count)`.
    """
    return ((tru - est) / tru) ** 2


def _make_gaussian_matrix(
    data_count: int,
    feature_count: int,
) -> mm.ndarray:
    """
    Create a matrix of i.i.d. Gaussian datapoints.

    Args:
        data_count:
            The number of data rows.
        feature_count:
            The number of data columns.

    Returns:
        An i.i.d. Gaussian matrix of shape `(data_count, feature_count)`.
    """
    return mm.array(np.random.randn(data_count, feature_count))


def _make_uniform_matrix(
    data_count: int,
    feature_count: int,
) -> mm.ndarray:
    """
    Create a matrix of i.i.d. Uniform(0,1) datapoints.

    Args:
        data_count:
            The number of data rows.
        feature_count:
            The number of data columns.

    Returns:
        An i.i.d. Gaussian matrix of shape `(data_count, feature_count)`.
    """
    return mm.array(np.random.rand(data_count, feature_count))


def _make_gaussian_dict(
    data_count: int,
    feature_count: int,
    response_count: int,
    categorical: bool = False,
) -> Dict[str, mm.ndarray]:
    """
    Create a data dict including "input", "output", and "labels" keys mapping to
    i.i.d. Gaussian matrices.

    Args:
        data_count:
            The number of data rows.
        feature_count:
            The number of data columns in the `data["input"]` matrix.
        resonse_count:
            The number of data columns in the `data["output"]` matrix.
        categorical:
            If `True`, convert the `data["output"]` matrix to a one-hot encoding
            matrix.

    Returns:
        A dict with keys `"input"` mapping to a `(data_count, feature_count)`
        matrix, `"output"` mapping to a `(data_count, response_count)` matrix,
        and `"labels"` mapping to a `(data_count)` vector.
    """
    locations = _make_gaussian_matrix(data_count, feature_count)
    observations = _make_gaussian_matrix(data_count, response_count)
    labels = mm.argmax(observations, axis=1)
    if categorical is True:
        observations = mm.eye(response_count)[labels] - (1 / response_count)
    return {
        "input": locations,
        "output": observations,
        "labels": labels,
    }


def _make_gaussian_data(
    train_count: int,
    test_count: int,
    feature_count: int,
    response_count: int,
    categorical: bool = False,
) -> Tuple[Dict[str, mm.ndarray], Dict[str, mm.ndarray]]:
    """
    Create train and test dicts including `"input"`, `"output"`, and `"labels"`
    keys mapping to i.i.d. Gaussian matrices.

    Args:
        train_count:
            The number of train rows.
        test_count:
            The number of test rows.
        feature_count:
            The number of data columns in the `test["input"]` and
            `train["input"]` matrices.
        resonse_count:
            The number of data columns in the `test["output"]` and
            `train["input"]` matrices.
        categorical:
            If `True`, convert the `data["output"]` matrix to a one-hot encoding
            matrix.

    Returns
    -------
    train:
        A dict with keys `"input"` mapping to a matrix of shape
        `(train_count, feature_count)`, `"output`" mapping to a matrix of shape
        `(train_count, response_count)`, and `"labels"` mapping to a vector of
        shape `(train_count,)`.
    test:
        A dict with keys `"input"` mapping to a matrix of shape
        `(test_count, feature_count)`, `"output`" mapping to a matrix of shape
        `(test_count, response_count)`, and `"labels"` mapping to a vector of
        shape `(test_count,)`.
    """
    return (
        _make_gaussian_dict(
            train_count, feature_count, response_count, categorical=categorical
        ),
        _make_gaussian_dict(
            test_count, feature_count, response_count, categorical=categorical
        ),
    )


def _subsample(
    data: Dict[str, mm.ndarray],
    sample_count: int,
) -> Dict[str, mm.ndarray]:
    """
    Randomly sample row indices without replacement from data dict.

    NOTICE: This function and its Dict data format are intended for testing.

    Args:
        data:
            A dict with keys `"input"` and `"output"`. `data["input"]` maps to a
            matrix of shape `(data_count, feature_count)` whose rows consist of
            observation vectors. `data["output"]` maps to a matrix of shape
            `(data_count, response_count)` whose rows consist of response
            vectors.
        sample_count:
            The desired number of samples.

    Returns:
        A dict of the same form as `data`, but containing only the sampled
        indices.
    """
    count = data["input"].shape[0]
    samples = mm.array(
        np.random.choice(count, sample_count, replace=False), dtype=mm.itype
    )
    return {
        "input": data["input"][samples, :],
        "output": data["output"][samples, :],
    }


def _balanced_subsample(
    data: Dict[str, mm.ndarray],
    sample_count: int,
) -> Dict[str, mm.ndarray]:
    """
    Randomly sample row indices without replacement from data dict, ensuring
    that classes receive as close to equal representation as possible.

    Partitions the data based upon their true classes, and attempts to randomly
    sample without replacement a balanced quantity within each partition. May
    not work well on heavily skewed data except with very small sample sizes.

    NOTICE: This function and its Dict data format are intended for testing.

    Args:
        data:
            A dict with keys `"input"` and `"output"`. `data["input"]` maps to a
            matrix of shape `(data_count, feature_count)` whose rows consist of
            observation vectors. `data["output"]` maps to a matrix of shape
            `(data_count, response_count)` whose rows consist of response
            vectors.
        sample_count:
            The desired number of samples.

    Returns:
        A dict of the same form as `data`, but containing only the sampled
        indices, who have as close to parity in class representation as
        possible.
    """
    labels = mm.argmax(data["output"], axis=1)
    classes = mm.unique(labels)
    class_count = len(classes)
    each_sample_count = int(sample_count / class_count)

    class_indices = [np.where(labels == i)[0] for i in classes]
    sample_sizes = [
        np.min((len(arr), each_sample_count)) for arr in class_indices
    ]

    balanced_samples = np.concatenate(
        [
            np.random.choice(class_indices[i], sample_sizes[i], replace=False)
            for i in range(class_count)
        ]
    )
    return {
        "input": mm.array(data["input"][balanced_samples, :]),
        "output": mm.array(data["output"][balanced_samples, :]),
    }


def _normalize(X: mm.ndarray) -> mm.ndarray:
    """
    Normalizes data matrix to have row l2-norms of 1

    Args:
        X:
            A matrix of shape `(data_count, feature_count)`.

    Returns:
        A row-normalized matrix of shape `(data-count, feature_count)`.
    """
    return X * mm.sqrt(X.shape[1] / mm.sum(X**2, axis=1))[:, None]


def _get_sigma_sq_series(
    K: mm.ndarray,
    nn_targets_column: mm.ndarray,
    eps: float,
) -> mm.ndarray:
    """
    Return the series of sigma^2 scale parameters for each neighborhood
    solve.

    NOTE[bwp]: This function is only for testing purposes.

    Args:
        K:
            A tensor of shape `(batch_count, nn_count, nn_count)` containing
            the `(nn_count, nn_count` -shaped kernel matrices corresponding
            to each of the batch elements.
        nn_targets:
            Tensor of floats of shape `(batch_count, nn_count, 1)` containing
            one dimension of the expected response for each nearest neighbor of
            each batch element.

    Returns:
        A vector of shape `(response_count)` listing the value of sigma^2
        for the given response dimension.
    """
    batch_count, nn_count, _ = nn_targets_column.shape

    sigmas = np.zeros((batch_count,))
    for i, el in enumerate(_get_sigma_sq(K, nn_targets_column, eps)):
        sigmas[i] = el
    return mm.array(sigmas / nn_count)


def _get_sigma_sq(
    K: mm.ndarray,
    nn_targets_column: mm.ndarray,
    eps: float,
) -> Generator[float, None, None]:
    """
    Generate series of :math:`\\sigma^2` scale parameters for each
    individual solve along a single dimension:

    .. math::
        \\sigma^2 = \\frac{1}{k} * Y_{nn}^T K_{nn}^{-1} Y_{nn}

    Here :math:`Y_{nn}` and :math:`K_{nn}` are the target and kernel
    matrices with respect to the nearest neighbor set in scope, where
    :math:`k` is the number of nearest neighbors.

    Args:
        K:
            A tensor of shape `(batch_count, nn_count, nn_count)` containing
            the `(nn_count, nn_count` -shaped kernel matrices corresponding
            to each of the batch elements.
        nn_targets_column:
            Tensor of floats of shape `(batch_count, nn_count, 1)` containing
            one dimension of the expected response for each nearest neighbor of
            each batch element.

    Return:
        A generator producing `batch_count` optimal values of
        :math:`\\sigma^2` for each neighborhood for the given response
        dimension.
    """
    batch_count, nn_count, _ = nn_targets_column.shape
    for j in range(batch_count):
        Y_0 = nn_targets_column[j, :, 0]
        yield Y_0 @ mm.linalg.solve(K[j, :, :] + eps * mm.eye(nn_count), Y_0)


def _check_ndarray(
    assert_fn: Callable,
    array: mm.ndarray,
    dtype: Type,
    ctype: Type = mm.ndarray,
    shape: Optional[Tuple[int, ...]] = None,
):
    assert_fn(type(array), ctype)
    assert_fn(array.dtype, dtype)
    if shape is not None:
        assert_fn(array.shape, shape)


def _precision_assert(assert_fn, *args, low_bound=4, high_bound=7):
    if config.state.ftype == "32":
        return assert_fn(*args, low_bound)
    else:
        return assert_fn(*args, high_bound)


def _consistent_assert(assert_fn, *args):
    """
    Performs an assert on the root core if in mpi, otherwise performs the assert
    as normal.

    The purpose of this function is to allow the existing serial testing harness
    to also test the mpi implementations without the need for additional codes.

    Args:
        assert_fn:
            An absl assert function.
        args:
            Arguments to the assert function
    """
    if _is_mpi_mode() is True:
        if config.mpi_state.comm_world.Get_rank() == 0:
            assert_fn(*args)
    else:
        assert_fn(*args)


def _make_heteroscedastic_test_nugget(
    batch_count: int, nn_count: int, eps_mag: float
):
    """
    Produces a test heteroscedastic 3D tensor parameter of shape
    `(batch_count, nn_count, nn_count)`.

    NOTE[amd]: This function is only for testing purposes.

    Args:
        batch_count:
            Number of points to be predicted.
        nn_count:
            Number of nearest neighbors in the kernel.
        eps:
            Maximum noise magnitude.


    Return:
        A `(batch_count, nn_count)` shaped tensor for heteroscedastic
        noise modeling.
    """
    eps_tensor = eps_mag * mm.ones((batch_count, nn_count))
    return eps_tensor



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
MuyGPs implementation
"""
import MuyGPyS._src.math as mm
from MuyGPyS._src.gp.muygps import (
    _muygps_fast_posterior_mean,
)


class FastPosteriorMean:
    def __init__(
        self,
    ):
        self._fn = _muygps_fast_posterior_mean

    def __call__(
        self,
        Kcross: mm.ndarray,
        coeffs_tensors: mm.ndarray,
    ) -> mm.ndarray:
        return self._fn(Kcross, coeffs_tensors)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from typing import Callable

from MuyGPyS.gp import MuyGPS
from MuyGPyS._src.optimize.chassis.numpy import (
    _scipy_optimize,
    _bayes_opt_optimize as _numpy_bayes_opt_optimize,
)


def _bayes_opt_optimize(
    muygps: MuyGPS,
    obj_fn: Callable,
    verbose: bool = False,
    **kwargs,
) -> MuyGPS:
    return _numpy_bayes_opt_optimize(muygps, obj_fn, verbose=verbose, **kwargs)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from .chassis import optimize_from_tensors



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Multivariate MuyGPs implementation
"""
from copy import deepcopy
import MuyGPyS._src.math as mm
from MuyGPyS._src.gp.muygps import _mmuygps_fast_posterior_mean
from MuyGPyS.gp.muygps import MuyGPS
from MuyGPyS.gp.mean import PosteriorMean
from MuyGPyS.gp.noise import HeteroscedasticNoise
from MuyGPyS.gp.noise.perturbation import select_perturb_fn
from MuyGPyS.gp.sigma_sq import SigmaSq
from MuyGPyS.gp.variance import PosteriorVariance


class MultivariateMuyGPS:
    """
    Multivariate Local Kriging Gaussian Process.

    Performs approximate GP inference by locally approximating an observation's
    response using its nearest neighbors with a separate kernel allocated for
    each response dimension, implemented as individual
    :class:`MuyGPyS.gp.muygps.MuyGPS` objects.

    This class is similar in interface to :class:`MuyGPyS.gp.muygps.MuyGPS`, but
    requires a list of hyperparameter dicts at initialization.

    Example:
        >>> from MuyGPyS.gp import MultivariateMuyGPS as MMuyGPS
        >>> k_kwargs1 = {
        ...         "eps": {"val": 1e-5},
        ...         "nu": {"val": 0.67, "bounds": (0.1, 2.5)},
        ...         "length_scale": {"val": 7.2},
        ... }
        >>> k_kwargs2 = {
        ...         "eps": {"val": 1e-5},
        ...         "nu": {"val": 0.38, "bounds": (0.1, 2.5)},
        ...         "length_scale": {"val": 7.2},
        ... }
        >>> k_args = [k_kwargs1, k_kwargs2]
        >>> mmuygps = MMuyGPS("matern", *k_args)

    We can realize kernel tensors for each of the models contained within a
    `MultivariateMuyGPS` object by iterating over its `models` member. Once we
    have computed `pairwise_diffs` and `crosswise_diffs` tensors, it
    is straightforward to perform each of these realizations.

    Example:
        >>> for model in MuyGPyS.models:
        >>>         K = model.kernel(pairwise_diffs)
        >>>         Kcross = model.kernel(crosswise_diffs)
        >>>         # do something with K and Kcross...

    Args
        model_args:
            Dictionaries defining each internal
            :class:`MuyGPyS.gp.muygps.MuyGPS` instance.
    """

    def __init__(
        self,
        *model_args,
    ):
        self.models = [MuyGPS(**args) for args in model_args]
        self.sigma_sq = SigmaSq(len(self.models))

    def fixed(self) -> bool:
        """
        Checks whether all kernel and model parameters are fixed for each model,
        excluding :math:`\\sigma^2`.

        Returns:
            Returns `True` if all parameters in all models are fixed, and
            `False` otherwise.
        """
        return bool(all([model.fixed() for model in self.models]))

    def posterior_mean(
        self,
        pairwise_diffs: mm.ndarray,
        crosswise_diffs: mm.ndarray,
        batch_nn_targets: mm.ndarray,
    ) -> mm.ndarray:
        """
        Performs simultaneous posterior mean inference on provided difference
        tensors and the target matrix.

        Computes parallelized local solves of systems of linear equations using
        the kernel realizations, one for each internal model, of the last two
        dimensions of `pairwise_diffs` along with `crosswise_diffs` and
        `batch_nn_targets` to predict responses in terms of the posterior mean.
        Assumes that difference tensors `pairwise_diffs` and `crosswise_diffs`
        are already computed and given as arguments.

        Returns the predicted response in the form of a posterior mean for each
        element of the batch of observations by solving a system of linear
        equations induced by each kernel functor, one per response dimension, in
        a generalization of Equation (3.4) of [muyskens2021muygps]_. For each
        batch element :math:`\\mathbf{x}_i` we compute

        .. math::
            \\widehat{Y}_{NN} (\\mathbf{x}_i \\mid X_{N_i})_{:,j} =
                K^{(j)}_\\theta (\\mathbf{x}_i, X_{N_i})
                (K^{(j)}_\\theta (X_{N_i}, X_{N_i}) + \\varepsilon_j I_k)^{-1}
                Y(X_{N_i})_{:,j}.

        Here :math:`X_{N_i}` is the set of nearest neighbors of
        :math:`\\mathbf{x}_i` in the training data, :math:`K^{(j)}_\\theta` is
        the kernel functor associated with the jth internal model, corresponding
        to the jth response dimension, :math:`\\varepsilon_j I_k` is a diagonal
        homoscedastic noise matrix whose diagonal is the value of the
        `self.models[j].eps` hyperparameter, and :math:`Y(X_{N_i})_{:,j}` is the
        `(batch_count,)` vector of the jth responses of the nearest neighbors
        given by a slice of the `batch_nn_targets` argument.

        Args:
            pairwise_diffs:
                A tensor of shape
                `(batch_count, nn_count, nn_count, feature_count)` containing
                the `(nn_count, nn_count, feature_count)`-shaped pairwise
                nearest neighbor difference tensors corresponding to each of the
                batch elements.
            crosswise_diffs:
                A matrix of shape `(batch_count, nn_count, feature_count)` whose
                rows list the difference between each feature of each batch
                element element and its nearest neighbors.
            batch_nn_targets:
                A tensor of shape `(batch_count, nn_count, response_count)`
                listing the vector-valued responses for the nearest neighbors
                of each batch element.

        Returns:
            A matrix of shape `(batch_count, response_count)` whose rows are
            the predicted response for each of the given indices.
        """
        batch_count, nn_count, response_count = batch_nn_targets.shape
        responses = mm.zeros((batch_count, response_count))
        for i, model in enumerate(self.models):
            K = model.kernel(pairwise_diffs)
            Kcross = model.kernel(crosswise_diffs)
            responses = mm.assign(
                responses,
                model.posterior_mean(
                    K,
                    Kcross,
                    batch_nn_targets[:, :, i].reshape(batch_count, nn_count, 1),
                ).reshape(batch_count),
                slice(None),
                i,
            )
        return responses

    def posterior_variance(
        self,
        pairwise_diffs: mm.ndarray,
        crosswise_diffs: mm.ndarray,
    ) -> mm.ndarray:
        """
        Performs simultaneous posterior variance inference on provided
        difference tensors.

        Return the local posterior variances of each prediction, corresponding
        to the diagonal elements of a covariance matrix. For each batch element
        :math:`\\mathbf{x}_i`, we compute

        .. math::
            Var(\\widehat{Y}_{NN} (\\mathbf{x}_i \\mid X_{N_i}))_j =
                K^{(j)}_\\theta (\\mathbf{x}_i, \\mathbf{x}_i) -
                K^{(j)}_\\theta (\\mathbf{x}_i, X_{N_i})
                (K^{(j)}_\\theta (X_{N_i}, X_{N_i}) + \\varepsilon I_k)^{-1}
                K^{(j)}_\\theta (X_{N_i}, \\mathbf{x}_i).

        Args:
            pairwise_diffs:
                A tensor of shape
                `(batch_count, nn_count, nn_count, feature_count)` containing
                the `(nn_count, nn_count, feature_count)`-shaped pairwise
                nearest neighbor difference tensors corresponding to each of the
                batch elements.
            crosswise_diffs:
                A matrix of shape `(batch_count, nn_count, feature_count)` whose
                rows list the difference between each feature of each batch
                element element and its nearest neighbors.

        Returns:
            A vector of shape `(batch_count, response_count)` consisting of the
            diagonal elements of the posterior variance for each model.
        """
        batch_count, _, _ = crosswise_diffs.shape
        response_count = len(self.models)
        diagonal_variance = mm.zeros((batch_count, response_count))
        for i, model in enumerate(self.models):
            K = model.kernel(pairwise_diffs)
            Kcross = model.kernel(crosswise_diffs)
            ss = self.sigma_sq()[i]
            diagonal_variance = mm.assign(
                diagonal_variance,
                model.posterior_variance(K, Kcross).reshape(batch_count) * ss,
                slice(None),
                i,
            )
        return diagonal_variance

    def fast_coefficients(
        self,
        pairwise_diffs_fast: mm.ndarray,
        train_nn_targets_fast: mm.ndarray,
    ) -> mm.ndarray:
        """
        Produces coefficient tensor for fast posterior mean inference given in
        Equation (8) of [dunton2022fast]_.

        To form the tensor, we compute

        .. math::
            \\mathbf{C}_{N^*}(i, :, j) =
                (K_{\\hat{\\theta_j}} (X_{N^*}, X_{N^*}) +
                \\varepsilon I_k)^{-1} Y(X_{N^*}).

        Here :math:`X_{N^*}` is the union of the nearest neighbor of the ith
        test point and the `nn_count - 1` nearest neighbors of this nearest
        neighbor, :math:`K_{\\hat{\\theta_j}}` is the trained kernel functor
        corresponding the jth response and specified by `self.models`,
        :math:`\\varepsilon I_k` is a diagonal homoscedastic noise matrix whose
        diagonal  is the value of the `self.eps` hyperparameter,
        and :math:`Y(X_{N^*})` is the `(train_count, response_count)`
        matrix of responses corresponding to the training features indexed
        by $N^*$.

        Args:
            pairwise_diffs:
                A tensor of shape
                `(train_count, nn_count, nn_count, feature_count)` containing
                the `(nn_count, nn_count, feature_count)`-shaped pairwise
                nearest neighbor difference tensors corresponding to each of the
                batch elements.
            batch_nn_targets:
                A tensor of shape `(train_count, nn_count, response_count)`
                listing the vector-valued responses for the nearest neighbors
                of each batch element.
        Returns:
            A tensor of shape `(batch_count, nn_count, response_count)`
            whose entries comprise the precomputed coefficients for fast
            posterior mean inference.
        """

        train_count, nn_count, response_count = train_nn_targets_fast.shape
        coeffs_tensor = mm.zeros((train_count, nn_count, response_count))

        for i, model in enumerate(self.models):
            K = model.kernel(pairwise_diffs_fast)
            perturb_fn = select_perturb_fn(model.eps)
            mm.assign(
                coeffs_tensor,
                model.fast_coefficients(
                    perturb_fn(K, model.eps()),
                    train_nn_targets_fast[:, :, i],
                ),
                slice(None),
                slice(None),
                i,
            )

        return coeffs_tensor

    def fast_posterior_mean(
        self,
        crosswise_diffs: mm.ndarray,
        coeffs_tensor: mm.ndarray,
    ) -> mm.ndarray:
        """
        Performs fast posterior mean inference using provided crosswise
        differences and precomputed coefficient matrix.

        Returns the predicted response in the form of a posterior
        mean for each element of the batch of observations, as computed in
        Equation (9) of [dunton2022fast]_. For each test point
        :math:`\\mathbf{z}`, we compute

        .. math::
            \\widehat{Y} (\\mathbf{z} \\mid X) =
                K_\\theta (\\mathbf{z}, X_{N^*}) \\mathbf{C}_{N^*}.

        Here :math:`X_{N^*}` is the union of the nearest neighbor of the queried
        test point :math:`\\mathbf{z}` and the nearest neighbors of that
        training point, :math:`K_\\theta` is the kernel functor specified by
        `self.kernel`, and :math:`\\mathbf{C}_{N^*}` is the matrix of
        precomputed coefficients given in Equation (8) of [dunton2022fast]_.

        Args:
            crosswise_diffs:
                A matrix of shape `(batch_count, nn_count, feature_count)` whose
                rows list the difference between each feature of each batch
                element element and its nearest neighbors.
            coeffs_tensor:
                A tensor of shape `(batch_count, nn_count, response_count)`
                providing the precomputed coefficients.

        Returns:
            A matrix of shape `(batch_count, response_count)` whose rows are
            the predicted response for each of the given indices.
        """
        Kcross = mm.zeros(coeffs_tensor.shape)
        for i, model in enumerate(self.models):
            mm.assign(
                Kcross,
                model.kernel(crosswise_diffs),
                slice(None),
                slice(None),
                i,
            )
        return _mmuygps_fast_posterior_mean(Kcross, coeffs_tensor)

    def apply_new_noise(self, new_noise):
        """
        Updates the heteroscedastic noise parameters of a MultivariateMuyGPs
        model.

        Args:
            new_noise:
                A matrix of shape
                `(test_count, nn_count, nn_count, response_count)` containing
                the measurement noise corresponding to the nearest neighbors
                of each test point and each response.
        Returns:
            A MultivariateMuyGPs model with updated heteroscedastic noise
            parameters.
        """
        ret = deepcopy(self)
        for i, model in enumerate(ret.models):
            model.eps = HeteroscedasticNoise(new_noise[:, :, :, i], "fixed")
            model._mean_fn = PosteriorMean(model.eps)
            model._var_fn = PosteriorVariance(model.eps, model.sigma_sq)
        return ret



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from typing import Callable

from MuyGPyS.gp import MuyGPS


def _scipy_optimize(
    muygps: MuyGPS,
    obj_fn: Callable,
    verbose: bool = False,
    **kwargs,
) -> MuyGPS:
    raise NotImplementedError(
        "Scipy optimization is not supported in MuyGPs PyTorch module."
    )


def _bayes_opt_optimize(
    muygps: MuyGPS,
    obj_fn: Callable,
    verbose: bool = False,
    **kwargs,
) -> MuyGPS:
    raise NotImplementedError(
        "Bayesian optimization is not supported in MuyGPs PyTorch module."
    )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Noise modeling

Defines data structures and functors that handle noise priors for MuyGPs models.
"""

from MuyGPyS.gp.hyperparameter import ScalarHyperparameter


class NullNoise(ScalarHyperparameter):
    """
    A zero noise assumption model.
    """

    def __init__(self, *args, **kwargs):
        self.val = 0.0
        self.bounds = "fixed"

    def __call__(self, *args, **kwargs):
        return 0.0



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from jax import jit

import MuyGPyS._src.math.jax as jnp


@jit
def _homoscedastic_perturb(K: jnp.ndarray, eps: float) -> jnp.ndarray:
    _, nn_count, _ = K.shape
    return K + eps * jnp.eye(nn_count)


@jit
def _heteroscedastic_perturb(K: jnp.ndarray, eps: jnp.ndarray) -> jnp.ndarray:
    batch_count, nn_count, _ = K.shape
    ret = K.copy()
    indices = (
        jnp.repeat(jnp.arange(batch_count), nn_count),
        jnp.tile(jnp.arange(nn_count), batch_count),
        jnp.tile(jnp.arange(nn_count), batch_count),
    )
    ret = ret.at[indices].add(eps.flatten())

    return ret



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from MuyGPyS._src.util import _collect_implementation

(
    _muygps_posterior_mean,
    _muygps_diagonal_variance,
    _muygps_fast_posterior_mean,
    _muygps_fast_posterior_mean_precompute,
    _mmuygps_fast_posterior_mean,
) = _collect_implementation(
    "MuyGPyS._src.gp.muygps",
    "_muygps_posterior_mean",
    "_muygps_diagonal_variance",
    "_muygps_fast_posterior_mean",
    "_muygps_fast_posterior_mean_precompute",
    "_mmuygps_fast_posterior_mean",
)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from scipy.special import softmax
from sklearn.metrics import log_loss

import MuyGPyS._src.math.numpy as np


def _cross_entropy_fn(
    predictions: np.ndarray,
    targets: np.ndarray,
    ll_eps: float = 1e-15,
) -> float:
    one_hot_targets = np.where(targets > 0.0, 1.0, 0.0)
    softmax_predictions = softmax(predictions, axis=1)
    return log_loss(
        one_hot_targets, softmax_predictions, eps=ll_eps, normalize=False
    )


def _mse_fn_unnormalized(
    predictions: np.ndarray,
    targets: np.ndarray,
) -> float:
    return np.sum((predictions - targets) ** 2)


def _mse_fn(
    predictions: np.ndarray,
    targets: np.ndarray,
) -> float:
    batch_count, response_count = predictions.shape
    return _mse_fn_unnormalized(predictions, targets) / (
        batch_count * response_count
    )


def _lool_fn_unscaled(
    predictions: np.ndarray, targets: np.ndarray, variances: np.ndarray
) -> float:
    return np.sum(
        np.divide((predictions - targets) ** 2, variances) + np.log(variances)
    )


def _lool_fn(
    predictions: np.ndarray,
    targets: np.ndarray,
    variances: np.ndarray,
    sigma_sq: np.ndarray,
) -> float:
    return _lool_fn_unscaled(
        predictions, targets, np.outer(variances, sigma_sq)
    )


def _pseudo_huber_fn(
    predictions: np.ndarray, targets: np.ndarray, boundary_scale: float = 1.5
) -> float:
    return boundary_scale**2 * np.sum(
        np.sqrt(1 + np.divide(targets - predictions, boundary_scale) ** 2) - 1
    )


def _looph_fn_unscaled(
    predictions: np.ndarray,
    targets: np.ndarray,
    variances: np.ndarray,
    boundary_scale: float = 1.5,
) -> float:
    boundary_scale_sq = boundary_scale**2
    return np.sum(
        boundary_scale_sq
        * np.sqrt(
            1
            + np.divide(targets - predictions, variances * boundary_scale) ** 2
        )
        - boundary_scale_sq
        + np.log(variances)
    )


def _looph_fn(
    predictions: np.ndarray,
    targets: np.ndarray,
    variances: np.ndarray,
    sigma_sq: np.ndarray,
    boundary_scale: float = 1.5,
) -> float:
    return _looph_fn_unscaled(
        predictions,
        targets,
        np.outer(variances, sigma_sq),
        boundary_scale=boundary_scale,
    )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from .embed import (
    apply_distortion,
    embed_with_distortion_model,
)
from .isotropic import IsotropicDistortion
from .null import NullDistortion
from .anisotropic import AnisotropicDistortion

from MuyGPyS._src.gp.tensors import _l2 as l2
from MuyGPyS._src.gp.tensors import _F2 as F2



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from MuyGPyS._src.util import _collect_implementation

(
    _rbf_fn,
    _matern_05_fn,
    _matern_15_fn,
    _matern_25_fn,
    _matern_inf_fn,
    _matern_gen_fn,
) = _collect_implementation(
    "MuyGPyS._src.gp.kernels",
    "_rbf_fn",
    "_matern_05_fn",
    "_matern_15_fn",
    "_matern_25_fn",
    "_matern_inf_fn",
    "_matern_gen_fn",
)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Convenience functions for optimizing the `sigma_sq` parameter of
:class:`~MuyGPyS.gp.muygps.MuyGPS` objects.

Currently only supports an analytic approximation, but will support other
methods in the future.
"""

from copy import deepcopy
from typing import Callable, Optional

import MuyGPyS._src.math as mm
from MuyGPyS.gp import MuyGPS, MultivariateMuyGPS as MMuyGPS
from MuyGPyS._src.optimize.sigma_sq import _analytic_sigma_sq_optim
from MuyGPyS.optimize.utils import _switch_on_sigma_method
from MuyGPyS.gp.noise.perturbation import select_perturb_fn


def muygps_sigma_sq_optim(
    muygps: MuyGPS,
    pairwise_diffs: mm.ndarray,
    nn_targets: mm.ndarray,
    sigma_method: Optional[str] = "analytic",
) -> MuyGPS:
    """
    Optimize the value of the :math:`\\sigma^2` scale parameter for each
    response dimension.

    The optimization to be applied depends upon the value of `sigma_method`.

    Args:
        muygps:
            The model to be optimized.
        pairwise_diffs:
            A tensor of shape `(batch_count, nn_count, nn_count, feature_count)`
            containing the `(nn_count, nn_count, feature_count)`-shaped pairwise
            nearest neighbor difference tensors corresponding to each of the
            batch elements.
        nn_targets:
            Tensor of floats of shape `(batch_count, nn_count, response_count)`
            containing the expected response for each nearest neighbor of each
            batch element.
        sigma_method:
            The optimization method to apply. Currently only supports
            `"analytic"` and `None`.

    Returns:
        A new MuyGPs model whose sigma_sq parameter has been optimized.
    """
    return _switch_on_sigma_method(
        sigma_method,
        muygps_analytic_sigma_sq_optim,
        lambda muygps, pairwise_diffs, nn_targets: muygps,
        muygps,
        pairwise_diffs,
        nn_targets,
    )


def mmuygps_sigma_sq_optim(
    mmuygps: MMuyGPS,
    pairwise_diffs: mm.ndarray,
    nn_targets: mm.ndarray,
    sigma_method: Optional[str] = "analytic",
) -> MMuyGPS:
    """
    Optimize the value of the :math:`\\sigma^2` scale parameter for each
    response dimension of a MultivariateMuyGPS object.

    The optimization to be applied depends upon the value of `sigma_method`.

    Args:
        mmuygps:
            The model to be optimized.
        pairwise_diffs:
            A tensor of shape `(batch_count, nn_count, nn_count, feature_count)`
            containing the `(nn_count, nn_count, feature_count)`-shaped pairwise
            nearest neighbor difference tensors corresponding to each of the
            batch elements.
        nn_targets:
            Tensor of floats of shape `(batch_count, nn_count, response_count)`
            containing the expected response for each nearest neighbor of each
            batch element.
        sigma_method:
            The optimization method to apply. Currently only supports
            `"analytic"` and `None`.

    Returns:
        A new MultivariateMuyGPs model whose sigma_sq parameter has been
        optimized.
    """
    return _switch_on_sigma_method(
        sigma_method,
        mmuygps_analytic_sigma_sq_optim,
        lambda mmuygps, pairwise_diffs, nn_targets: mmuygps,
        mmuygps,
        pairwise_diffs,
        nn_targets,
    )


def make_sigma_sq_optim(
    sigma_method: Optional[str], muygps: MuyGPS
) -> Callable:
    return _switch_on_sigma_method(
        sigma_method,
        make_analytic_sigma_sq_optim,
        make_none_sigma_sq_optim,
        muygps,
        _analytic_sigma_sq_optim,
        select_perturb_fn(muygps.eps),
    )


def make_none_sigma_sq_optim(muygps: MuyGPS, *args) -> Callable:
    return lambda: lambda K, nn_targets, **kwargs: muygps.sigma_sq()


def make_analytic_sigma_sq_optim(
    muygps: MuyGPS, analytic_optim_fn, perturb_fn: Callable
) -> Callable:
    if not muygps.eps.fixed():

        def ss_opt_fn(K, nn_targets, **kwargs):
            return analytic_optim_fn(perturb_fn(K, kwargs["eps"]), nn_targets)

    else:

        def ss_opt_fn(K, nn_targets, **kwargs):
            return analytic_optim_fn(perturb_fn(K, muygps.eps()), nn_targets)

    return ss_opt_fn


def muygps_analytic_sigma_sq_optim(
    muygps: MuyGPS, pairwise_diffs: mm.ndarray, nn_targets: mm.ndarray
) -> MuyGPS:
    """
    Optimize the value of the :math:`\\sigma^2` scale parameter for each
    response dimension.

    We approximate :math:`\\sigma^2` by way of averaging over the analytic
    solution from each local kernel.

    .. math::
        \\sigma^2 = \\frac{1}{bk} * \\sum_{i \\in B}
                    Y_{nn_i}^T K_{nn_i}^{-1} Y_{nn_i}

    Here :math:`Y_{nn_i}` and :math:`K_{nn_i}` are the target and kernel
    matrices with respect to the nearest neighbor set in scope, where
    :math:`k` is the number of nearest neighbors and :math:`b = |B|` is the
    number of batch elements considered.

    Args:
        muygps:
            The model to be optimized.
        pairwise_diffs:
            A tensor of shape `(batch_count, nn_count, nn_count, feature_count)`
            containing the `(nn_count, nn_count, feature_count)`-shaped pairwise
            nearest neighbor difference tensors corresponding to each of the
            batch elements.
        nn_targets:
            Tensor of floats of shape `(batch_count, nn_count, response_count)`
            containing the expected response for each nearest neighbor of each
            batch element.

    Returns:
        A new MuyGPs model whose sigma_sq parameter has been optimized.
    """
    ret = deepcopy(muygps)
    perturb_fn = select_perturb_fn(ret.eps)
    K = ret.kernel(pairwise_diffs)
    ss = _analytic_sigma_sq_optim(perturb_fn(K, ret.eps()), nn_targets)
    ret.sigma_sq._set(ss)
    return ret


def mmuygps_analytic_sigma_sq_optim(
    mmuygps: MMuyGPS, pairwise_diffs: mm.ndarray, nn_targets: mm.ndarray
) -> MMuyGPS:
    """
    Optimize the value of the :math:`\\sigma^2` scale parameter for each
    response dimension.

    We approximate :math:`\\sigma^2` by way of averaging over the analytic
    solution from each local kernel.

    .. math::
        \\sigma^2 = \\frac{1}{bk} * \\sum_{i \\in B}
                    Y_{nn_i}^T K_{nn_i}^{-1} Y_{nn_i}

    Here :math:`Y_{nn_i}` and :math:`K_{nn_i}` are the target and kernel
    matrices with respect to the nearest neighbor set in scope, where
    :math:`k` is the number of nearest neighbors and :math:`b = |B|` is the
    number of batch elements considered.

    Args:
        muygps:
            The model to be optimized.
        pairwise_diffs:
            A tensor of shape `(batch_count, nn_count, nn_count, feature_count)`
            containing the `(nn_count, nn_count, feature_count)`-shaped pairwise
            nearest neighbor difference tensors corresponding to each of the
            batch elements.
        nn_targets:
            Tensor of floats of shape `(batch_count, nn_count, response_count)`
            containing the expected response for each nearest neighbor of each
            batch element.

    Returns:
        A new MuyGPs model whose sigma_sq parameter has been optimized.
    """
    ret = deepcopy(mmuygps)
    batch_count, nn_count, response_count = nn_targets.shape
    if response_count != len(ret.models):
        raise ValueError(
            f"Response count ({response_count}) does not match the number "
            f"of models ({len(ret.models)})."
        )

    sigma_sqs = mm.zeros((response_count,))
    for i, model in enumerate(ret.models):
        perturb_fn = select_perturb_fn(model.eps)
        K = model.kernel(pairwise_diffs)
        new_sigma_val = _analytic_sigma_sq_optim(
            perturb_fn(K, model.eps()),
            nn_targets[:, :, i].reshape(batch_count, nn_count, 1),
        )[0]
        sigma_sqs = mm.assign(sigma_sqs, new_sigma_val, i)
        model.sigma_sq._set(mm.atleast_1d(sigma_sqs[i]))
    ret.sigma_sq._set(sigma_sqs)
    return ret



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from mpi4py import MPI

import MuyGPyS._src.math.numpy as np
from MuyGPyS import config
from MuyGPyS._src.optimize.sigma_sq.numpy import (
    _analytic_sigma_sq_optim_unnormalized,
)

world = config.mpi_state.comm_world
# rank = config.mpi_state.comm_world.Get_rank()
# size = config.mpi_state.comm_world.Get_size()


def _analytic_sigma_sq_optim(
    K: np.ndarray,
    nn_targets: np.ndarray,
) -> np.ndarray:
    local_batch_count, nn_count, _ = nn_targets.shape
    local_sum = _analytic_sigma_sq_optim_unnormalized(K, nn_targets)
    global_sum = world.allreduce(local_sum, op=MPI.SUM)
    global_batch_count = world.allreduce(local_batch_count, op=MPI.SUM)
    return global_sum / (nn_count * global_batch_count)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT
import MuyGPyS._src.math.numpy as np
from MuyGPyS._src.gp.muygps.numpy import (
    _muygps_posterior_mean,
    _muygps_diagonal_variance,
)


def _muygps_fast_posterior_mean(
    Kcross: np.ndarray,
    coeffs_tensor: np.ndarray,
) -> np.ndarray:
    raise NotImplementedError(
        'Function "muygps_fast_posterior_mean" does not support mpi!'
    )


def _mmuygps_fast_posterior_mean(
    Kcross: np.ndarray,
    coeffs_tensor: np.ndarray,
) -> np.ndarray:
    raise NotImplementedError(
        'Function "mmuygps_fast_posterior_mean" does not support mpi!'
    )


def _muygps_fast_posterior_mean_precompute(
    K: np.ndarray,
    eps: float,
    train_nn_targets_fast: np.ndarray,
) -> np.ndarray:
    raise NotImplementedError(
        'Function "muygps_fast_posterior_mean_precompute" does not support mpi!'
    )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Resources and high-level API for a deep kernel learning with MuyGPs.

:func:`~MuyGPyS.examples.muygps_torch.train_deep_kernel_muygps` is a high-level
API for training deep kernel MuyGPs models for regression.

:func:`~MuyGPyS.examples.muygps_torch.predict_model` is a high-level API for
generating predictions at test locations given a trained model.
"""

from typing import Dict

from torch.optim.lr_scheduler import ExponentialLR

import torch
import numpy as np
from MuyGPyS import config
from MuyGPyS.gp.tensors import (
    pairwise_tensor,
    crosswise_tensor,
)
from MuyGPyS.optimize.loss import lool_fn_unscaled as lool_fn
from MuyGPyS.neighbors import NN_Wrapper

from MuyGPyS.torch import MultivariateMuyGPs_layer

if config.state.backend != "torch":
    import warnings

    warnings.warn(
        f"torch-only code cannot be run in {config.state.backend} mode"
    )

if config.state.ftype != "32":
    import warnings

    warnings.warn(
        "torch optimization only supports 32-bit floats, not "
        f"{config.state.ftype}-bit"
    )


mse_loss = torch.nn.MSELoss()
l1_loss = torch.nn.L1Loss()
bce_loss = torch.nn.BCELoss()
ce_loss = torch.nn.CrossEntropyLoss()


def predict_single_model(
    model,
    test_features: torch.Tensor,
    train_features: torch.Tensor,
    train_responses: torch.Tensor,
    nbrs_lookup: NN_Wrapper,
    nn_count: int,
):
    """
    Generate predictions using a PyTorch model containing at least one
    `MuyGPyS.torch.muygps_layer.MuyGPs_layer` in its structure. Note that
    the custom PyTorch MuyGPs_layer objects only support the Matern kernel.
    Support for more kernels will be added in future releases.

    Args:
        model:
            A custom PyTorch.nn.Module object containing an
            `embedding` component and one
            `MuyGPyS.torch.muygps_layer.MuyGPs_layer` layer.
        test_features:
            A torch.Tensor of shape `(test_count, feature_count)` containing
            the test features to be regressed.
        train_features:
            A torch.Tensor of shape `(train_count, feature_count)` containing
            the training features.
        train_responses:
            A torch.Tensor of shape `(train_count, response_count)` containing
            the training responses corresponding to each feature.
        nbrs_lookup:
            A NN_Wrapper nearest neighbor lookup data structure.

    Returns
    -------
    predictions:
        A torch.Tensor of shape `(test_count, response_count)` whose rows are
        the predicted response for each of the given test feature.
    variances:
        A torch.Tensor of shape `(batch_count,response_count)` shape consisting
        of the diagonal elements of the posterior variance.
    """
    if model.embedding is None:
        raise NotImplementedError("MuyGPs PyTorch model requires embedding.")
    train_features_embedded = model.embedding(train_features).detach().numpy()
    test_features_embedded = model.embedding(test_features).detach().numpy()

    test_count = test_features_embedded.shape[0]

    nn_indices_test, _ = nbrs_lookup._get_nns(
        test_features_embedded, nn_count=nn_count
    )

    train_features_embedded = torch.from_numpy(train_features_embedded)
    test_features_embedded = torch.from_numpy(test_features_embedded)

    test_nn_targets = train_responses[nn_indices_test, :]

    crosswise_diffs = crosswise_tensor(
        test_features_embedded,
        train_features_embedded,
        torch.arange(test_count),
        nn_indices_test,
    )

    pairwise_diffs = pairwise_tensor(train_features_embedded, nn_indices_test)

    Kcross = model.GP_layer.muygps_model.kernel(crosswise_diffs)
    K = model.GP_layer.muygps_model.kernel(pairwise_diffs)

    predictions = model.GP_layer.muygps_model.posterior_mean(
        K, Kcross, test_nn_targets
    )

    variances = model.GP_layer.muygps_model.posterior_variance(K, Kcross)

    return predictions, variances


def predict_multiple_model(
    model,
    test_features: torch.Tensor,
    train_features: torch.Tensor,
    train_responses: torch.Tensor,
    nbrs_lookup: NN_Wrapper,
    nn_count: int,
):
    """
    Generate predictions using a PyTorch model containing a
    `MuyGPyS.torch.muygps_layer.MultivariateMuyGPs_layer` in its structure.
    Meant for the case in which there is more than one GP model used to model
    multiple outputs. Note that the custom PyTorch MultivariateMuyGPs_layer
    objects only support the Matern kernel. Support for more kernels will be
    added in future releases.

    Args:
        model:
            A custom PyTorch.nn.Module object containing an
            `embedding` component and one
            `MuyGPyS.torch.muygps_layer.MultivariateMuyGPs_layer` layer.
        test_features:
            A torch.Tensor of shape `(test_count, feature_count)` containing
            the test features to be regressed.
        train_features:
            A torch.Tensor of shape `(train_count, feature_count)` containing
            the training features.
        train_responses:
            A torch.Tensor of shape `(train_count, response_count)` containing
            the training responses corresponding to each feature.
        nbrs_lookup:
            A NN_Wrapper nearest neighbor lookup data structure.

    Returns
    -------
    predictions:
        A torch.Tensor of shape `(test_count, response_count)` whose rows are
        the predicted response for each of the given test feature.
    variances:
        A torch.Tensor of shape `(batch_count,)` consisting of the diagonal
        elements of the posterior variance, or a matrix of shape
        `(batch_count, response_count)` for a multidimensional response.
    """
    if model.embedding is None:
        raise NotImplementedError("MuyGPs PyTorch model requires embedding.")

    train_features_embedded = model.embedding(train_features).detach().numpy()
    test_features_embedded = model.embedding(test_features).detach().numpy()

    test_count = test_features_embedded.shape[0]

    nn_indices_test, _ = nbrs_lookup._get_nns(
        test_features_embedded, nn_count=nn_count
    )

    train_features_embedded = torch.from_numpy(train_features_embedded)
    test_features_embedded = torch.from_numpy(test_features_embedded)

    test_nn_targets = train_responses[nn_indices_test, :]

    crosswise_diffs = crosswise_tensor(
        test_features_embedded,
        train_features_embedded,
        torch.arange(test_count),
        nn_indices_test,
    )

    pairwise_diffs = pairwise_tensor(train_features_embedded, nn_indices_test)

    (
        _,
        nn_count,
        response_count,
    ) = model.batch_nn_targets.shape

    Kcross = torch.zeros(test_count, nn_count, response_count)
    K = torch.zeros(test_count, nn_count, nn_count, response_count)

    for i, muygps_model in enumerate(
        model.GP_layer.multivariate_muygps_model.models
    ):
        Kcross[:, :, i] = muygps_model.kernel(crosswise_diffs)
        K[:, :, :, i] = muygps_model.kernel(pairwise_diffs)

    predictions = model.GP_layer.multivariate_muygps_model.posterior_mean(
        K, Kcross, test_nn_targets
    )

    variances = model.GP_layer.multivariate_muygps_model.posterior_variance(
        K, Kcross
    )

    return predictions, variances


def predict_model(
    model,
    test_features: torch.Tensor,
    train_features: torch.Tensor,
    train_responses: torch.Tensor,
    nbrs_lookup: NN_Wrapper,
    nn_count: int,
):
    """
    Generate predictions using a PyTorch model containing a
    `MuyGPyS.torch.muygps_layer.MuyGPs_layer` layer or a
    `MuyGPyS.torch.muygps_layer.MultivariateMuyGPs_layer` layer in its
    structure. Note that the custom PyTorch layers for MuyGPs
    objects only support the Matern kernel. Support for more kernels will be
    added in future releases.

    Example:
        >>> #model must be defined as a PyTorch model inheriting from
        ... #torch.nn.Module. Must have two components: model.embedding
        ... #(e.g., a neural net) and another component model.GP_layer.
        >>> from MuyGPyS.testing.test_utils import _make_gaussian_data
        >>> from MuyGPyS.neighbors import NN_Wrapper
        >>> train, test = _make_gaussian_data(10000, 1000, 100, 10)
        >>> nn_count = 10
        >>> nbrs_lookup = NN_Wrapper(train['input'], nn_count, nn_method="hnsw")
        >>> predictions, variances = predict_model(
        ... model,
        ... torch.from_numpy(test['input']),
        ... torch.from_numpy(train['input']),
        ... torch.from_numpy(train['output']),
        ... nbrs_lookup,
        ... nn_count)



    Args:
        model:
            A custom PyTorch.nn.Module object containing an embedding component
            and one MuyGPs_layer or MultivariateMuyGPS_layer layer.
        test_features:
            A torch.Tensor of shape `(test_count, feature_count)` containing
            the test features to be regressed.
        train_features:
            A torch.Tensor of shape `(train_count, feature_count)` containing
            the training features.
        train_responses:
            A torch.Tensor of shape `(train_count, response_count)` containing
            the training responses corresponding to each feature.
        nbrs_lookup:
            A NN_Wrapper nearest neighbor lookup data structure.

    Returns
    -------
    predictions:
        A torch.Tensor of shape `(test_count, response_count)` whose rows are
        the predicted response for each of the given test feature.
    variances:
        A torch.Tensor of shape `(batch_count,)` consisting of the diagonal
        elements of the posterior variance, or a matrix of shape
        `(batch_count, response_count)` for a multidimensional response.
    """
    if model.GP_layer is None:
        raise NotImplementedError("MuyGPs PyTorch model requires GP_layer.")
    if isinstance(model.GP_layer, MultivariateMuyGPs_layer):
        return predict_multiple_model(
            model,
            test_features,
            train_features,
            train_responses,
            nbrs_lookup,
            nn_count,
        )
    else:
        return predict_single_model(
            model,
            test_features,
            train_features,
            train_responses,
            nbrs_lookup,
            nn_count,
        )


def train_deep_kernel_muygps(
    model,
    train_features: torch.Tensor,
    train_responses: torch.Tensor,
    batch_indices: torch.Tensor,
    nbrs_lookup: NN_Wrapper,
    training_iterations=10,
    optimizer_method=torch.optim.Adam,
    learning_rate=1e-3,
    scheduler_decay=0.95,
    loss_function="lool",
    update_frequency=1,
    verbose=False,
    nn_kwargs: Dict = dict(),
):
    """
    Train a PyTorch model containing an embedding component and
    a `MuyGPyS.torch.muygps_layer.MuyGPs_layer` layer or a
    `MuyGPyS.torch.muygps_layer. MultivariateMuyGPs_layer` layer in its
    structure. Note that the custom PyTorch layers for MuyGPs models only
    support the Matern kernel. Support for more kernels will be added in
    future releases.

    Example:
        >>> #model must be defined as a PyTorch model inheriting from
        ... #torch.nn.Module. Must have two components: model.embedding
        ... #(e.g., a neural net) and another component model.GP_layer.
        >>> from MuyGPyS.testing.test_utils import _make_gaussian_data
        >>> from MuyGPyS.neighbors import NN_Wrapper
        >>> from MuyGPyS.examples.muygps_torch import train_deep_kernel_muygps
        >>> from MuyGPyS._src.optimize.loss import _lool_fn as lool_fn
        >>> train, test = _make_gaussian_data(10000, 1000, 100, 10)
        >>> nn_count = 10
        >>> nbrs_lookup = NN_Wrapper(train['input'], nn_count, nn_method="hnsw")
        >>> batch_count = 100
        >>> train_count = 10000
        >>> batch_indices, batch_nn_indices = sample_batch(nbrs_lookup,
        ... batch_count, train_count)
        >>> nbrs_struct, model_trained = train_deep_kernel_muygps(
        ... model=model,
        ... train_features=torch.from_numpy(train['input']),
        ... train_responses=torch.from_numpy(train['output']),
        ... batch_indices=torch.from_numpy(batch_indices),
        ... nbrs_lookup=nbrs_lookup,
        ... training_iterations=10,
        ... optimizer_method=torch.optim.Adam,
        ... learning_rate=1e-3,
        ... scheduler_decay=0.95,
        ... loss_function=lool_fn,
        ... update_frequency=1)

    Args:
        model:
            A custom PyTorch.nn.Module object containing at least one
            embedding layer and one MuyGPs_layer or MultivariateMuyGPS_layer
            layer.
        train_features:
            A torch.Tensor of shape `(train_count, feature_count)` containing
            the training features.
        train_responses:
            A torch.Tensor of shape `(train_count, response_count)` containing
            the training responses corresponding to each feature.
        batch_indices:
            A torch.Tensor of shape `(batch_count,)` containing the indices of
            the training batch.
        nbrs_lookup:
            A NN_Wrapper nearest neighbor lookup data structure.
        training_iterations:
            The number of training iterations to be used in training.
        optimizer method:
            An optimization method from the torch.optim class.
        learning_rate:
            The learning rate to be applied during training.
        schedule_decay:
            The exponential decay rate to be applied to the learning rate.
        loss function:
            The loss function to be used in training. Defaults to "lool" for
            leave-one-out likelihood. Other options are "mse" for
            mean-squared error, "ce" for cross entropy loss, "bce" for binary
            cross entropy loss, and "l1" for L1 loss.
        update_frequency:
            Tells the training procedure how frequently the nearest neighbor
            structure should be updated. An update frequency of n indicates that
            every n epochs the nearest neighbor structure should be updated.
        verbose:
            Indicates whether or not to include print statements during
            training.
        nn_kwargs:
            Parameters for the nearest neighbors wrapper. See
            :class:`MuyGPyS.neighbors.NN_Wrapper` for the supported methods and
            their parameters.

    Returns
    -------
    nbrs_lookup:
        A NN_Wrapper object containing the nearest neighbors of the embedded
        training data.
    model:
        A trained deep kernel MuyGPs model.
    """
    if model.embedding is None:
        raise NotImplementedError("MuyGPs PyTorch model requires embedding.")
    optimizer = optimizer_method(
        [
            {"params": model.parameters()},
        ],
        lr=learning_rate,
    )
    scheduler = ExponentialLR(optimizer, gamma=scheduler_decay)
    nn_count = nbrs_lookup.nn_count
    batch_features = train_features[batch_indices, :]
    batch_responses = train_responses[batch_indices, :]

    loss_function = loss_function.lower()

    if loss_function == "mse":
        loss_func = mse_loss
    elif loss_function == "bce":
        loss_func = bce_loss
    elif loss_function == "ce":
        loss_func = ce_loss
    elif loss_function == "lool":
        loss_func = lool_fn
    else:
        raise ValueError(f"loss function {loss_function} is not supported")

    for i in range(training_iterations):
        model.train()
        optimizer.zero_grad()
        predictions, variances = model(train_features)

        if loss_function == "lool":
            loss = loss_func(
                predictions.squeeze(),
                batch_responses.squeeze(),
                variances.squeeze(),
            )
        else:
            loss = loss_func(predictions, batch_responses)
        loss.sum().backward()
        optimizer.step()
        scheduler.step()

        if np.mod(i, update_frequency) == 0:
            if verbose is True:
                print(
                    "Iter %d/%d - Loss: %.10f"
                    % (i + 1, training_iterations, loss.sum().item())
                )
            model.eval()
            nbrs_lookup = NN_Wrapper(
                model.embedding(train_features).detach().numpy(),
                nn_count,
                **nn_kwargs,
            )
            batch_nn_indices, _ = nbrs_lookup._get_nns(
                model.embedding(batch_features).detach().numpy(),
                nn_count=nn_count,
            )
            batch_nn_targets = train_responses[batch_nn_indices, :]

            model.batch_nn_indices = batch_nn_indices
            model.batch_nn_targets = batch_nn_targets

        torch.cuda.empty_cache()
    nbrs_lookup = NN_Wrapper(
        model.embedding(train_features).detach().numpy(),
        nn_count,
        **nn_kwargs,
    )
    batch_nn_indices, _ = nbrs_lookup._get_nns(
        model.embedding(batch_features).detach().numpy(),
        nn_count=nn_count,
    )
    batch_nn_targets = train_responses[batch_nn_indices, :]
    model.batch_nn_indices = batch_nn_indices
    model.batch_nn_targets = batch_nn_targets
    return nbrs_lookup, model


def update_nearest_neighbors(
    model,
    train_features: torch.Tensor,
    train_responses: torch.Tensor,
    batch_indices: torch.Tensor,
    nn_count: int,
    nn_kwargs: Dict = dict(),
):
    """
    Update the nearest neighbors after deformation via a PyTorch model
    containing an embedding component and a
    `MuyGPyS.torch.muygps_layer.MuyGPs_layer` layer or a
    `MuyGPyS.torch.muygps_layer. MultivariateMuyGPs_layer` layer in its
    structure.

    Example:
        >>> #model must be defined as a PyTorch model inheriting from
        ... #torch.nn.Module. Must have two components: model.embedding
        ... #(e.g., a neural net) and another component model.GP_layer.
        >>> from MuyGPyS.testing.test_utils import _make_gaussian_data
        >>> from MuyGPyS.neighbors import NN_Wrapper
        >>> from MuyGPyS.examples.muygps_torch import update_nearest_neighbors
        >>> train, test = _make_gaussian_data(10000, 1000, 100, 10)
        >>> nn_count = 10
        >>> batch_count = 100
        >>> train_count = 10000
        >>> batch_indices, batch_nn_indices = sample_batch(nbrs_lookup, batch_count, train_count)
        >>> nbrs_struct, model_trained = update_nearest_neighbors(
        ... model=model,
        ... train_features=torch.from_numpy(train['input']),
        ... train_responses=torch.from_numpy(train['output']),
        ... batch_indices=torch.from_numpy(batch_indices),
        ... nn_count=nn_count,)

    Args:
        model:
            A custom PyTorch.nn.Module object containing at least one
            embedding layer and one MuyGPs_layer or MultivariateMuyGPS_layer
            layer.
        train_features:
            A torch.Tensor of shape `(train_count, feature_count)` containing
            the training features.
        train_responses:
            A torch.Tensor of shape `(train_count, response_count)` containing
            the training responses corresponding to each feature.
        batch_indices:
            A torch.Tensor of shape `(batch_count,)` containing the indices of
            the training batch.
        nn_count:
            A torch.int64 giving the number of nearest neighbors.
        nn_kwargs:
            Parameters for the nearest neighbors wrapper. See
            :class:`MuyGPyS.neighbors.NN_Wrapper` for the supported methods and
            their parameters.

    Returns
    -------
    nbrs_lookup:
        A NN_Wrapper object containing the updated nearest neighbors of the
        embedded training data.
    model:
        A deep kernel MuyGPs model with updated nearest neighbors.
    """
    if model.embedding is None:
        raise NotImplementedError("MuyGPs PyTorch model requires embedding.")
    batch_features = train_features[batch_indices, :]
    nbrs_lookup = NN_Wrapper(
        model.embedding(train_features).detach().numpy(),
        nn_count,
        **nn_kwargs,
    )
    batch_nn_indices, _ = nbrs_lookup._get_nns(
        model.embedding(batch_features).detach().numpy(),
        nn_count=nn_count,
    )
    batch_nn_targets = train_responses[batch_nn_indices, :]
    model.batch_nn_indices = batch_nn_indices
    model.batch_nn_targets = batch_nn_targets
    return nbrs_lookup, model



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from jax import jit

import MuyGPyS._src.math.jax as jnp


@jit
def _muygps_posterior_mean(
    K: jnp.ndarray,
    Kcross: jnp.ndarray,
    batch_nn_targets: jnp.ndarray,
    **kwargs,
) -> jnp.ndarray:
    batch_count, nn_count, response_count = batch_nn_targets.shape
    responses = Kcross.reshape(batch_count, 1, nn_count) @ jnp.linalg.solve(
        K, batch_nn_targets
    )
    return responses.reshape(batch_count, response_count)


@jit
def _muygps_diagonal_variance(
    K: jnp.ndarray,
    Kcross: jnp.ndarray,
    **kwargs,
) -> jnp.ndarray:
    batch_count, nn_count = Kcross.shape
    return 1 - jnp.sum(
        Kcross
        * jnp.linalg.solve(K, Kcross.reshape(batch_count, nn_count, 1)).reshape(
            batch_count, nn_count
        ),
        axis=1,
    )


@jit
def _muygps_fast_posterior_mean(
    Kcross: jnp.ndarray,
    coeffs_tensor: jnp.ndarray,
) -> jnp.ndarray:
    return jnp.einsum("ij,ijk->ik", Kcross, coeffs_tensor)


@jit
def _mmuygps_fast_posterior_mean(
    Kcross: jnp.ndarray,
    coeffs_tensor: jnp.ndarray,
) -> jnp.ndarray:
    return jnp.einsum("ijk,ijk->ik", Kcross, coeffs_tensor)


@jit
def _muygps_fast_posterior_mean_precompute(
    K: jnp.ndarray,
    train_nn_targets_fast: jnp.ndarray,
) -> jnp.ndarray:
    return jnp.linalg.solve(K, train_nn_targets_fast)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from MuyGPyS._src.math.torch import nn
from MuyGPyS.torch import MuyGPs_layer, MultivariateMuyGPs_layer


class SVDKMuyGPs(nn.Module):
    def __init__(
        self,
        muygps_model,
        batch_indices,
        batch_nn_indices,
        batch_targets,
        batch_nn_targets,
    ):
        super().__init__()
        self.embedding = nn.Sequential(
            nn.Linear(40, 30),
            nn.Dropout(0.5),
            nn.PReLU(1),
            nn.Linear(30, 10),
            nn.Dropout(0.5),
            nn.PReLU(1),
        )
        self.muygps_model = muygps_model
        self.batch_indices = batch_indices
        self.batch_nn_indices = batch_nn_indices
        self.batch_targets = batch_targets
        self.batch_nn_targets = batch_nn_targets
        self.GP_layer = MuyGPs_layer(
            self.muygps_model,
            self.batch_indices,
            self.batch_nn_indices,
            self.batch_targets,
            self.batch_nn_targets,
        )

    def forward(self, x):
        predictions = self.embedding(x)
        predictions, variances = self.GP_layer(predictions)
        return predictions, variances


class SVDKMultivariateMuyGPs(nn.Module):
    def __init__(
        self,
        multivariate_muygps_model,
        batch_indices,
        batch_nn_indices,
        batch_targets,
        batch_nn_targets,
    ):
        super().__init__()
        self.embedding = nn.Sequential(
            nn.Linear(40, 30),
            nn.Dropout(0.5),
            nn.ELU(1),
            nn.Linear(30, 10),
            nn.Dropout(0.5),
            nn.ELU(1),
        )
        self.multivariate_muygps_model = multivariate_muygps_model
        self.batch_indices = batch_indices
        self.batch_nn_indices = batch_nn_indices
        self.batch_targets = batch_targets
        self.batch_nn_targets = batch_nn_targets
        self.GP_layer = MultivariateMuyGPs_layer(
            self.multivariate_muygps_model,
            self.batch_indices,
            self.batch_nn_indices,
            self.batch_targets,
            self.batch_nn_targets,
        )

    def forward(self, x):
        predictions = self.embedding(x)
        predictions, variances = self.GP_layer(predictions)
        return predictions, variances



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Copy of JAX config object.

This code is included so that MuyGPySConfig can inherit from JaxConfig if the
JAX dependency is not included.
"""

# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# TODO(phawkins): this file triggers a pytype bug.
# pytype: skip-file

import contextlib
import itertools
import os
import sys
import threading
from typing import Any, List, Callable, NamedTuple, Optional
import warnings

# from jax._src import lib
# from jax._src.lib import jax_jit
# from jax._src.lib import transfer_guard_lib


def bool_env(varname: str, default: bool) -> bool:
    """Read an environment variable and interpret it as a boolean.
    True values are (case insensitive): 'y', 'yes', 't', 'true', 'on', and '1';
    false values are 'n', 'no', 'f', 'false', 'off', and '0'.
    Args:
      varname: the name of the variable
      default: the default boolean value
    Raises: ValueError if the environment variable is anything else.
    """
    val = os.getenv(varname, str(default))
    val = val.lower()
    if val in ("y", "yes", "t", "true", "on", "1"):
        return True
    elif val in ("n", "no", "f", "false", "off", "0"):
        return False
    else:
        raise ValueError(
            "invalid truth value %r for environment %r" % (val, varname)
        )


def int_env(varname: str, default: int) -> int:
    """Read an environment variable and interpret it as an integer."""
    return int(os.getenv(varname, str(default)))


UPGRADE_BOOL_HELP = (
    " This will be enabled by default in future versions of JAX, at which "
    "point all uses of the flag will be considered deprecated (following "
    "the `API compatibility policy "
    "<https://jax.readthedocs.io/en/latest/api_compatibility.html>`_)."
)

UPGRADE_BOOL_EXTRA_DESC = " (transient)"


class Config:
    _HAS_DYNAMIC_ATTRIBUTES = True

    def __init__(self):
        self.values = {}
        self.meta = {}
        self.FLAGS = NameSpace(self.read, self.update)
        self.use_absl = False
        self._contextmanager_flags = set()
        self._update_hooks = {}

        self.omnistaging_enabled = True  # TODO(mattjj): remove this

    def update(self, name, val):
        if self.use_absl:
            setattr(self.absl_flags.FLAGS, name, val)
        else:
            self.check_exists(name)
            if name not in self.values:
                raise Exception("Unrecognized config option: {}".format(name))
            self.values[name] = val

        hook = self._update_hooks.get(name, None)
        if hook:
            hook(val)

    def read(self, name):
        if name in self._contextmanager_flags:
            raise AttributeError(
                "For flags with a corresponding contextmanager, read their value "
                f"via e.g. `config.{name}` rather than `config.FLAGS.{name}`."
            )
        return self._read(name)

    def _read(self, name):
        if self.use_absl:
            return getattr(self.absl_flags.FLAGS, name)
        else:
            self.check_exists(name)
            return self.values[name]

    def add_option(
        self, name, default, opt_type, meta_args, meta_kwargs, update_hook=None
    ):
        if name in self.values:
            raise Exception("Config option {} already defined".format(name))
        self.values[name] = default
        self.meta[name] = (opt_type, meta_args, meta_kwargs)
        if update_hook:
            self._update_hooks[name] = update_hook
            update_hook(default)

    def check_exists(self, name):
        if name not in self.values:
            raise AttributeError("Unrecognized config option: {}".format(name))

    def DEFINE_bool(self, name, default, *args, **kwargs):
        update_hook = kwargs.pop("update_hook", None)
        self.add_option(
            name, default, bool, args, kwargs, update_hook=update_hook
        )

    def DEFINE_integer(self, name, default, *args, **kwargs):
        update_hook = kwargs.pop("update_hook", None)
        self.add_option(
            name, default, int, args, kwargs, update_hook=update_hook
        )

    def DEFINE_string(self, name, default, *args, **kwargs):
        update_hook = kwargs.pop("update_hook", None)
        self.add_option(
            name, default, str, args, kwargs, update_hook=update_hook
        )

    def DEFINE_enum(self, name, default, *args, **kwargs):
        update_hook = kwargs.pop("update_hook", None)
        self.add_option(
            name, default, "enum", args, kwargs, update_hook=update_hook
        )

    def config_with_absl(self):
        # Run this before calling `app.run(main)` etc
        import absl.flags as absl_FLAGS  # noqa: F401
        from absl import app, flags as absl_flags

        self.use_absl = True
        self.absl_flags = absl_flags
        absl_defs = {
            bool: absl_flags.DEFINE_bool,
            int: absl_flags.DEFINE_integer,
            str: absl_flags.DEFINE_string,
            "enum": absl_flags.DEFINE_enum,
        }

        for name, val in self.values.items():
            flag_type, meta_args, meta_kwargs = self.meta[name]
            absl_defs[flag_type](name, val, *meta_args, **meta_kwargs)
        app.call_after_init(lambda: self.complete_absl_config(absl_flags))

    def complete_absl_config(self, absl_flags):
        for name, _ in self.values.items():
            self.update(name, getattr(absl_flags.FLAGS, name))

    def parse_flags_with_absl(self):
        global already_configured_with_absl
        if not already_configured_with_absl:
            # Extract just the --jax... flags (before the first --) from argv. In some
            # environments (e.g. ipython/colab) argv might be a mess of things
            # parseable by absl and other junk.
            jax_argv = itertools.takewhile(lambda a: a != "--", sys.argv)
            jax_argv = ["", *(a for a in jax_argv if a.startswith("--jax"))]

            import absl.flags

            self.config_with_absl()
            absl.flags.FLAGS(jax_argv, known_only=True)
            self.complete_absl_config(absl.flags)
            already_configured_with_absl = True

            if not FLAGS.jax_omnistaging:
                raise Exception(
                    "Disabling of omnistaging is no longer supported in JAX version 0.2.12 and higher: "
                    "see https://github.com/google/jax/blob/main/design_notes/omnistaging.md.\n"
                    "To remove this warning, unset the JAX_OMNISTAGING environment variable."
                )

    def enable_omnistaging(self):
        warnings.warn(
            "enable_omnistaging() is a no-op in JAX versions 0.2.12 and higher;\n"
            "see https://github.com/google/jax/blob/main/design_notes/omnistaging.md"
        )

    def disable_omnistaging(self):
        raise Exception(
            "Disabling of omnistaging is no longer supported in JAX version 0.2.12 and higher: "
            "see https://github.com/google/jax/blob/main/design_notes/omnistaging.md."
        )

    def define_bool_state(
        self,
        name: str,
        default: bool,
        help: str,
        *,
        update_global_hook: Optional[Callable[[bool], None]] = None,
        update_thread_local_hook: Optional[
            Callable[[Optional[bool]], None]
        ] = None,
        upgrade: bool = False,
        extra_description: str = "",
    ):
        """Set up thread-local state and return a contextmanager for managing it.
        This function is a convenience wrapper. It defines a flag, environment
        variable, and corresponding thread-local state, which can be managed via the
        contextmanager it returns.
        The thread-local state value can be read via the ``config.<option_name>``
        attribute, where ``config`` is the singleton ``Config`` instance.
        Args:
          name: string, converted to lowercase to define the name of the config
            option (and absl flag). It is converted to uppercase to define the
            corresponding shell environment variable.
          default: boolean, a default value for the option.
          help: string, used to populate the flag help information as well as the
            docstring of the returned context manager.
          update_global_hook: a optional callback that is called with the updated
            value of the global state when it is altered or set initially.
          update_thread_local_hook: a optional callback that is called with the
            updated value of the thread-local state when it is altered or set
            initially.
          upgrade: optional indicator that this flag controls a canonical feature
            upgrade, so that it is `True` for the incoming functionality, `False`
            for the outgoing functionality to be deprecated.
          extra_description: string, optional: extra information to add to the
            summary description.
        Returns:
          A contextmanager to control the thread-local state value.
        Example:
          enable_foo = config.define_bool_state(
              name='jax_enable_foo',
              default=False,
              help='Enable foo.')
          # Now the JAX_ENABLE_FOO shell environment variable and --jax_enable_foo
          # command-line flag can be used to control the process-level value of
          # the configuration option, in addition to using e.g.
          # ``config.update("jax_enable_foo", True)`` directly. We can also use a
          # context manager:
          with enable_foo(True):
            ...
        The value of the thread-local state or flag can be accessed via
        ``config.jax_enable_foo``. Reading it via ``config.FLAGS.jax_enable_foo`` is
        an error.
        """
        name = name.lower()
        if upgrade:
            help += " " + UPGRADE_BOOL_HELP
            extra_description += UPGRADE_BOOL_EXTRA_DESC
        self.DEFINE_bool(
            name,
            bool_env(name.upper(), default),
            help,
            update_hook=update_global_hook,
        )
        self._contextmanager_flags.add(name)

        def get_state(self):
            val = getattr(_thread_local_state, name, unset)
            return val if val is not unset else self._read(name)

        setattr(Config, name, property(get_state))

        return _StateContextManager(
            name,
            help,
            update_thread_local_hook,
            extra_description=extra_description,
            default_value=True,
        )

    def define_enum_state(
        self,
        name: str,
        enum_values: List[str],
        default: Optional[str],
        help: str,
        update_global_hook: Optional[Callable[[str], None]] = None,
        update_thread_local_hook: Optional[
            Callable[[Optional[str]], None]
        ] = None,
    ):
        """Set up thread-local state and return a contextmanager for managing it.
        Args:
          name: string, converted to lowercase to define the name of the config
            option (and absl flag). It is converted to uppercase to define the
            corresponding shell environment variable.
          enum_values: list of strings representing the possible values for the
            option.
          default: optional string, default value.
          help: string, used to populate the flag help information as well as the
            docstring of the returned context manager.
        Returns:
          A contextmanager to control the thread-local state value.
        See docstring for ``define_bool_state``.
        """
        name = name.lower()
        default = os.getenv(name.upper(), default)
        if default is not None and default not in enum_values:
            raise ValueError(f'Invalid value "{default}" for JAX flag {name}')
        self.DEFINE_enum(
            name,
            default,
            enum_values=enum_values,
            help=help,
            update_hook=update_global_hook,
        )
        self._contextmanager_flags.add(name)

        def get_state(self):
            val = getattr(_thread_local_state, name, unset)
            return val if val is not unset else self._read(name)

        setattr(Config, name, property(get_state))

        def validate(new_val):
            if new_val is not None and (
                type(new_val) is not str or new_val not in enum_values
            ):
                raise ValueError(
                    f"new enum value must be None or in {enum_values}, "
                    f"got {new_val} of type {type(new_val)}."
                )

        return _StateContextManager(
            name, help, update_thread_local_hook, validate
        )

    def define_string_state(
        self,
        name: str,
        default: Optional[str],
        help: str,
        update_global_hook: Optional[Callable[[str], None]] = None,
        update_thread_local_hook: Optional[
            Callable[[Optional[str]], None]
        ] = None,
    ):
        """Set up thread-local state and return a contextmanager for managing it.
        See docstring for ``define_bool_state``.
        Args:
          name: string, converted to lowercase to define the name of the config
            option (and absl flag). It is converted to uppercase to define the
            corresponding shell environment variable.
          default: string, a default value for the option.
          help: string, used to populate the flag help information as well as the
            docstring of the returned context manager.
          update_global_hook: an optional callback that is called with the updated
            value of the global state when it is altered or set initially.
          update_thread_local_hook: an optional callback that is called with the
            updated value of the thread-local state when it is altered or set
            initially.
        Returns:
          A contextmanager to control the thread-local state value.
        """
        name = name.lower()
        default = os.getenv(name.upper(), default)
        self.DEFINE_string(
            name, default, help=help, update_hook=update_global_hook
        )
        self._contextmanager_flags.add(name)

        def get_state(self):
            val = getattr(_thread_local_state, name, unset)
            return val if val is not unset else self._read(name)

        setattr(Config, name, property(get_state))

        def validate(new_val):
            if new_val is not None and not isinstance(new_val, str):
                raise ValueError(
                    f"new string config value must be None or of type str,"
                    f" got {new_val} of type {type(new_val)}."
                )

        return _StateContextManager(
            name, help, update_thread_local_hook, validate
        )

    def _trace_context(self):
        """Returns a tuple of configuration values that affect tracing.
        These values are included in the cache key for linear_util.cache.
        Values included in this set should also most likely be included in
        the C++ JIT state, which is handled separately."""
        return (
            self.x64_enabled,
            self.jax_numpy_rank_promotion,
            self.jax_default_matmul_precision,
            self.jax_dynamic_shapes,
        )


class NoDefault:
    pass


no_default = NoDefault()


class _StateContextManager:
    def __init__(
        self,
        name,
        help,
        update_thread_local_hook,
        validate_new_val_hook: Optional[Callable[[Any], None]] = None,
        extra_description: str = "",
        default_value: Any = no_default,
    ):
        self._name = name
        self.__name__ = name[4:] if name.startswith("jax_") else name
        self.__doc__ = (
            f"Context manager for `{name}` config option"
            f"{extra_description}.\n\n{help}"
        )
        self._update_thread_local_hook = update_thread_local_hook
        self._validate_new_val_hook = validate_new_val_hook
        self._default_value = default_value

    @contextlib.contextmanager
    def __call__(self, new_val=no_default):
        if new_val is no_default:
            if self._default_value is not no_default:
                new_val = (
                    self._default_value
                )  # default_value provided to constructor
            else:
                # no default_value provided to constructor and no value provided as an
                # argument, so we raise an error
                raise TypeError(
                    f"Context manager for {self.__name__} config option "
                    "requires an argument representing the new value for "
                    "the config option."
                )
        if self._validate_new_val_hook:
            self._validate_new_val_hook(new_val)
        prev_val = getattr(_thread_local_state, self._name, unset)
        setattr(_thread_local_state, self._name, new_val)
        if self._update_thread_local_hook:
            self._update_thread_local_hook(new_val)
        try:
            yield
        finally:
            if prev_val is unset:
                delattr(_thread_local_state, self._name)
                if self._update_thread_local_hook:
                    self._update_thread_local_hook(None)
            else:
                setattr(_thread_local_state, self._name, prev_val)
                if self._update_thread_local_hook:
                    self._update_thread_local_hook(prev_val)

    def _add_hooks(self, update_global_hook, update_thread_local_hook):
        """Private method that adds hooks to an existing context-manager.
        Used to avoid cyclic import dependencies."""
        self._update_thread_local_hook = update_thread_local_hook
        config._update_hooks[self._name] = update_global_hook
        update_global_hook(config._read(self._name))


_thread_local_state = threading.local()


class _Unset:
    pass


unset = _Unset()


class NameSpace:
    def __init__(self, getter, setter):
        # must use super because we override this class's __setattr__, see
        # https://docs.python.org/3/reference/datamodel.html#object.__setattr__
        super().__setattr__("_getter", getter)
        super().__setattr__("_setter", setter)

    def __getattr__(self, name):
        return self._getter(name)

    def __setattr__(self, name, val):
        self._setter(name, val)


config = Config()
flags = config
FLAGS = flags.FLAGS

already_configured_with_absl = False


# The C++ JIT maintains its own copy of several configuration items as
# a global/thread-local state. These methods allow updates to part of the
# state when a configuration value changes.


class GlobalJitState(NamedTuple):
    numpy_rank_promotion: Optional[str] = None
    default_matmul_precision: Optional[Any] = None
    dynamic_shapes: bool = False


# def update_global_jit_state(**kw):
#     gs = jax_jit.global_state()
#     context = gs.extra_jit_context or GlobalJitState()
#     gs.extra_jit_context = context._replace(**kw)


class ThreadLocalJitState(NamedTuple):
    dynamic_trace_state: Optional[Any] = None
    numpy_rank_promotion: Optional[str] = None
    default_matmul_precision: Optional[Any] = None
    dynamic_shapes: bool = False


# def update_thread_local_jit_state(**kw):
#     tls = jax_jit.thread_local_state()
#     context = tls.extra_jit_context or ThreadLocalJitState()
#     tls.extra_jit_context = context._replace(**kw)


# # TODO(mattjj): remove all uses of this flag
# flags.DEFINE_bool(
#     "jax_omnistaging",
#     bool_env("JAX_OMNISTAGING", True),
#     help=(
#         "Deprecated. Setting this flag to False raises an error. Setting it "
#         "to True has no effect."
#     ),
# )

# flags.DEFINE_integer(
#     "jax_tracer_error_num_traceback_frames",
#     int_env("JAX_TRACER_ERROR_NUM_TRACEBACK_FRAMES", 5),
#     help="Set the number of stack frames in JAX tracer error messages.",
# )

# flags.DEFINE_bool(
#     "jax_pprint_use_color",
#     bool_env("JAX_PPRINT_USE_COLOR", False),
#     help="Enable jaxpr pretty-printing with colorful syntax highlighting.",
# )

# flags.DEFINE_bool(
#     "jax_host_callback_inline",
#     bool_env("JAX_HOST_CALLBACK_INLINE", False),
#     help="Inline the host_callback, if not in a staged context.",
# )
# flags.DEFINE_integer(
#     "jax_host_callback_max_queue_byte_size",
#     int_env("JAX_HOST_CALLBACK_MAX_QUEUE_BYTE_SIZE", int(256 * 1e6)),
#     help=(
#         "The size in bytes of the buffer used to hold outfeeds from each "
#         "device. When this capacity is reached consuming outfeeds from the "
#         "device is paused, thus potentially pausing the device computation, "
#         "until the Python callback consume more outfeeds."
#     ),
#     lower_bound=int(16 * 1e6),
# )
# flags.DEFINE_bool(
#     "jax_host_callback_outfeed",
#     bool_env("JAX_HOST_CALLBACK_OUTFEED", False),
#     help=(
#         "Use outfeed implementation for host_callback, even on CPU and GPU. "
#         "If false, use the CustomCall implementation. "
#         "Has no effect on TPU, since only the outfeed mechanism is implemented."
#     ),
# )
# flags.DEFINE_bool(
#     "jax_host_callback_ad_transforms",
#     bool_env("JAX_HOST_CALLBACK_AD_TRANSFORMS", False),
#     help=(
#         "Enable support for jvp/vjp for the host_callback primitives. Default is "
#         "False, which means that host_callback operates only on primals. "
#         "The flag exists only temporarily, for backward compatibility."
#     ),
# )

# # TODO(b/214340779): remove flag when XLA:CPU is improved.
# jax2tf_associative_scan_reductions = config.define_bool_state(
#     name="jax2tf_associative_scan_reductions",
#     default=False,
#     help=(
#         "JAX has two separate lowering rules for the cumulative reduction "
#         "primitives (cumsum, cumprod, cummax, cummin). On CPUs and GPUs it uses "
#         "a lax.associative_scan, while for TPUs it uses the HLO ReduceWindow. "
#         "The latter has a slow implementation on CPUs and GPUs. "
#         "By default, jax2tf uses the TPU lowering. Set this flag to True to "
#         "use the associative scan lowering usage, and only if it makes a difference "
#         "for your application. "
#         "See the jax2tf README.md for more details."
#     ),
# )

# enable_checks = config.define_bool_state(
#     name="jax_enable_checks",
#     default=False,
#     help="Turn on invariant checking for JAX internals. Makes things slower.",
# )

# check_tracer_leaks = config.define_bool_state(
#     name="jax_check_tracer_leaks",
#     default=False,
#     help=(
#         "Turn on checking for leaked tracers as soon as a trace completes. "
#         "Enabling leak checking may have performance impacts: some caching "
#         "is disabled, and other overheads may be added. Additionally, be aware "
#         "that some Python debuggers can cause false positives, so it is recommended "
#         "to disable any debuggers while leak checking is enabled."
#     ),
# )
# checking_leaks = functools.partial(check_tracer_leaks, True)

# debug_nans = config.define_bool_state(
#     name="jax_debug_nans",
#     default=False,
#     help=(
#         "Add nan checks to every operation. When a nan is detected on the "
#         "output of a jit-compiled computation, call into the un-compiled "
#         "version in an attempt to more precisely identify the operation "
#         "which produced the nan."
#     ),
# )

# debug_infs = config.define_bool_state(
#     name="jax_debug_infs",
#     default=False,
#     help=(
#         "Add inf checks to every operation. When an inf is detected on the "
#         "output of a jit-compiled computation, call into the un-compiled "
#         "version in an attempt to more precisely identify the operation "
#         "which produced the inf."
#     ),
# )

# log_compiles = config.define_bool_state(
#     name="jax_log_compiles",
#     default=False,
#     help=(
#         "Log a message each time every time `jit` or `pmap` compiles an XLA "
#         "computation. Logging is performed with `absl.logging`. When this "
#         "option is set, the log level is WARNING; otherwise the level is "
#         "DEBUG."
#     ),
# )

# parallel_functions_output_gda = config.define_bool_state(
#     name="jax_parallel_functions_output_gda",
#     default=False,
#     help="If True, pjit will output GSDAs.",
# )


# distributed_debug = config.define_bool_state(
#     name="jax_distributed_debug",
#     default=False,
#     help=(
#         "Enable logging useful for debugging multi-process distributed "
#         "computations. Logging is performed with `absl.logging` at WARNING "
#         "level."
#     ),
# )


# enable_custom_prng = config.define_bool_state(
#     name="jax_enable_custom_prng",
#     default=False,
#     upgrade=True,
#     help=(
#         "Enables an internal upgrade that allows one to define custom "
#         "pseudo-random number generator implementations."
#     ),
# )

# default_prng_impl = config.define_enum_state(
#     name="jax_default_prng_impl",
#     enum_values=["threefry2x32", "rbg", "unsafe_rbg"],
#     default="threefry2x32",
#     help=(
#         "Select the default PRNG implementation, used when one is not "
#         "explicitly provided at seeding time."
#     ),
# )

# enable_custom_vjp_by_custom_transpose = config.define_bool_state(
#     name="jax_enable_custom_vjp_by_custom_transpose",
#     default=False,
#     upgrade=True,
#     help=(
#         "Enables an internal upgrade that implements `jax.custom_vjp` by "
#         "reduction to `jax.custom_jvp` and `jax.custom_transpose`."
#     ),
# )

# hlo_source_file_canonicalization_regex = config.define_string_state(
#     name="jax_hlo_source_file_canonicalization_regex",
#     default=None,
#     help=(
#         "Used to canonicalize the source_path metadata of HLO instructions "
#         "by removing the given regex. If set, re.sub() is called on each "
#         "source_file with the given regex, and all matches are removed. "
#         "This can be used to avoid spurious cache misses when using the "
#         "persistent compilation cache, which includes HLO metadata in the "
#         "cache key."
#     ),
# )

# config.define_enum_state(
#     name="jax_default_dtype_bits",
#     enum_values=["32", "64"],
#     default="64",
#     help=(
#         "Specify bit width of default dtypes, either 32-bit or 64-bit. "
#         "This is a temporary flag that will be used during the process "
#         "of deprecating the ``jax_enable_x64`` flag."
#     ),
# )


# def _update_x64_global(val):
#     lib.jax_jit.global_state().enable_x64 = val


# def _update_x64_thread_local(val):
#     lib.jax_jit.thread_local_state().enable_x64 = val


# enable_x64 = config.define_bool_state(
#     name="jax_enable_x64",
#     default=False,
#     help="Enable 64-bit types to be used",
#     update_global_hook=_update_x64_global,
#     update_thread_local_hook=_update_x64_thread_local,
# )

# # TODO(phawkins): remove after fixing users of FLAGS.x64_enabled.
# config._contextmanager_flags.remove("jax_enable_x64")

# Config.x64_enabled = Config.jax_enable_x64  # type: ignore


# def _update_disable_jit_global(val):
#     lib.jax_jit.global_state().disable_jit = val


# def _update_disable_jit_thread_local(val):
#     lib.jax_jit.thread_local_state().disable_jit = val


# disable_jit = config.define_bool_state(
#     name="jax_disable_jit",
#     default=False,
#     help=("Disable JIT compilation and just call original Python."),
#     update_global_hook=_update_disable_jit_global,
#     update_thread_local_hook=_update_disable_jit_thread_local,
# )


# numpy_rank_promotion = config.define_enum_state(
#     name="jax_numpy_rank_promotion",
#     enum_values=["allow", "warn", "raise"],
#     default="allow",
#     help=(
#         "Control NumPy-style automatic rank promotion broadcasting "
#         '("allow", "warn", or "raise").'
#     ),
#     update_global_hook=lambda val: update_global_jit_state(
#         numpy_rank_promotion=val
#     ),
#     update_thread_local_hook=lambda val: update_thread_local_jit_state(
#         numpy_rank_promotion=val
#     ),
# )

# default_matmul_precision = config.define_enum_state(
#     name="jax_default_matmul_precision",
#     enum_values=["bfloat16", "tensorfloat32", "float32"],
#     default=None,
#     help=(
#         "Control the default matmul and conv precision for 32bit inputs.\n\n"
#         "Some platforms, like TPU, offer configurable precision levels for "
#         "matrix multiplication and convolution computations, trading off "
#         "accuracy for speed. The precision can be controlled for each "
#         "operation; for example, see the :func:`jax.lax.conv_general_dilated` "
#         "and :func:`jax.lax.dot` docstrings. But it can be useful to control "
#         "the default behavior obtained when an operation is not given a "
#         "specific precision.\n\n"
#         "This option can be used to control the default precision "
#         "level for computations involved in matrix multiplication and "
#         "convolution on 32bit inputs. The levels roughly describe the "
#         "precision at which scalar products are computed. The 'bfloat16' "
#         "option is the fastest and least precise; 'float32' is similar to "
#         "full float32 precision; 'tensorfloat32' is intermediate.\n\n"
#     ),
#     update_global_hook=lambda val: update_global_jit_state(
#         default_matmul_precision=val
#     ),
#     update_thread_local_hook=lambda val: update_thread_local_jit_state(
#         default_matmul_precision=val
#     ),
# )

# traceback_filtering = config.define_enum_state(
#     name="jax_traceback_filtering",
#     enum_values=["off", "tracebackhide", "remove_frames", "auto"],
#     default="auto",
#     help="Controls how JAX filters internal frames out of tracebacks.\n\n"
#     "Valid values are:\n"
#     ' * "off": disables traceback filtering.\n'
#     ' * "auto": use "tracebackhide" if running under a sufficiently '
#     'new IPython, or "remove_frames" otherwise.\n'
#     ' * "tracebackhide": adds "__tracebackhide__" annotations to '
#     " hidden stack frames, which some traceback printers support.\n"
#     ' * "remove_frames": removes hidden frames from tracebacks, and adds '
#     " the unfiltered traceback as a __cause__ of the exception.\n",
# )

# # This flag is temporary and for internal use.
# # TODO(tianjianlu): Removes after providing the information in BCOO meta data.
# bcoo_cusparse_lowering = config.define_bool_state(
#     name="jax_bcoo_cusparse_lowering",
#     default=False,
#     help=("Enables lowering BCOO ops to cuSparse."),
# )

# # TODO(mattjj): remove this flag when we ensure we only succeed at trace-staging
# # if the intended backend can handle lowering the result
# config.define_bool_state(
#     name="jax_dynamic_shapes",
#     default=False,
#     help=(
#         "Enables experimental features for staging out computations with "
#         "dynamic shapes."
#     ),
#     update_global_hook=lambda val: update_global_jit_state(dynamic_shapes=val),
#     update_thread_local_hook=lambda val: update_thread_local_jit_state(
#         dynamic_shapes=val
#     ),
# )

# config.define_bool_state(
#     name="jax_experimental_name_stack",
#     default=False,
#     help="Enable using the context manager-based name stack.",
# )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from jax import jit

import MuyGPyS._src.math.jax as jnp


@jit
def _analytic_sigma_sq_optim_unnormalized(
    K: jnp.ndarray,
    nn_targets: jnp.ndarray,
) -> jnp.ndarray:
    return jnp.sum(
        jnp.einsum("ijk,ijk->ik", nn_targets, jnp.linalg.solve(K, nn_targets)),
        axis=0,
    )


@jit
def _analytic_sigma_sq_optim(
    K: jnp.ndarray,
    nn_targets: jnp.ndarray,
) -> jnp.ndarray:
    batch_count, nn_count, _ = nn_targets.shape
    return _analytic_sigma_sq_optim_unnormalized(K, nn_targets) / (
        batch_count * nn_count
    )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from typing import Callable, Union, Dict, Optional

import MuyGPyS._src.math as mm

from MuyGPyS.gp.hyperparameter import ScalarHyperparameter
from MuyGPyS.gp.hyperparameter.experimental import (
    HierarchicalNonstationaryHyperparameter,
)
from MuyGPyS.gp.distortion.anisotropic import AnisotropicDistortion
from MuyGPyS.gp.distortion.isotropic import IsotropicDistortion
from MuyGPyS.gp.distortion.null import NullDistortion


def apply_distortion(distortion_fn: Callable, **length_scales):
    def distortion_applier(fn: Callable):
        def distorted_fn(diffs, batch_features=None, *args, **kwargs):
            inner_kwargs = {
                key: _optional_invoke_param(
                    kwargs[key], batch_features=batch_features
                )
                for key in kwargs
                if key.startswith("length_scale") and "_knot" not in key
            }
            for ls in length_scales:
                inner_kwargs.setdefault(
                    ls,
                    _optional_invoke_param(
                        length_scales[ls],
                        batch_features=batch_features,
                        **kwargs,
                    ),
                )
            outer_kwargs = {
                key: _optional_invoke_param(
                    kwargs[key], batch_features=batch_features
                )
                for key in kwargs
                if not key.startswith("length_scale")
            }
            return fn(
                distortion_fn(diffs, **inner_kwargs), *args, **outer_kwargs
            )

        return distorted_fn

    return distortion_applier


def _optional_invoke_param(
    param: Union[
        ScalarHyperparameter, HierarchicalNonstationaryHyperparameter, float
    ],
    batch_features: Optional[mm.ndarray] = None,
    **kwargs,
) -> float:
    if isinstance(param, ScalarHyperparameter):
        return param()
    if isinstance(param, HierarchicalNonstationaryHyperparameter):
        return param(batch_features, **kwargs)
    return param


def embed_with_distortion_model(
    fn: Callable,
    distortion_fn: Callable,
    length_scale: Union[
        ScalarHyperparameter,
        HierarchicalNonstationaryHyperparameter,
        Dict[str, ScalarHyperparameter],
        Dict[str, HierarchicalNonstationaryHyperparameter],
    ],
):
    if isinstance(length_scale, ScalarHyperparameter):
        length_scale = {"length_scale": length_scale}
    if isinstance(length_scale, HierarchicalNonstationaryHyperparameter):
        length_scale = {"length_scale": length_scale}
    if isinstance(distortion_fn, AnisotropicDistortion) or isinstance(
        distortion_fn, IsotropicDistortion
    ):
        return apply_distortion(distortion_fn, **length_scale)(fn)
    elif isinstance(distortion_fn, NullDistortion):
        return fn
    else:
        raise ValueError(f"Noise model {type(distortion_fn)} is not supported!")



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from MuyGPyS._src.util import _collect_implementation

(
    all,
    allclose,
    arange,
    argmax,
    array,
    atleast_1d,
    atleast_2d,
    assign,
    corrcoef,
    cov,
    cholesky,
    eye,
    exp,
    diagonal,
    divide,
    iarray,
    inf,
    int32,
    int64,
    itype,
    isclose,
    farray,
    float32,
    float64,
    ftype,
    full,
    linalg,
    linspace,
    log,
    logical_or,
    max,
    median,
    min,
    ndarray,
    ones,
    outer,
    parameter,
    prod,
    repeat,
    reshape,
    sqrt,
    sum,
    tile,
    unique,
    vstack,
    where,
    zeros,
) = _collect_implementation(
    "MuyGPyS._src.math",
    "all",
    "allclose",
    "arange",
    "argmax",
    "array",
    "atleast_1d",
    "atleast_2d",
    "assign",
    "corrcoef",
    "cov",
    "cholesky",
    "eye",
    "exp",
    "diagonal",
    "divide",
    "iarray",
    "inf",
    "int32",
    "int64",
    "itype",
    "isclose",
    "farray",
    "float32",
    "float64",
    "ftype",
    "full",
    "linalg",
    "linspace",
    "log",
    "logical_or",
    "max",
    "median",
    "min",
    "ndarray",
    "ones",
    "outer",
    "parameter",
    "prod",
    "repeat",
    "reshape",
    "sqrt",
    "sum",
    "tile",
    "unique",
    "vstack",
    "where",
    "zeros",
)


def promote(x):
    if isinstance(x, ndarray):
        return x
    else:
        return array(x)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from .homoscedastic import HomoscedasticNoise
from .heteroscedastic import HeteroscedasticNoise
from .null import NullNoise
from .perturbation import (
    noise_perturb,
    perturb_with_noise_model,
)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Noise modeling

Defines data structures and functors that handle noise priors for MuyGPs models.
"""

from typing import Tuple, Union

from MuyGPyS.gp.hyperparameter import ScalarHyperparameter


class HomoscedasticNoise(ScalarHyperparameter):
    """
    A scalar :math:`\\eps` prior noise parameter.

    :math:`\\epse` is a homoscedastic noise parameter used to build the "nugget"
    with the prior assumption that all observations are subject to i.i.d.
    Gaussian noise. Can be set at initialization time or left subject to
    optimization, in which case (positive) bounds are specified.

    Args:
        val:
            A positive scalar, or the strings `"sample"` or `"log_sample"`.
        bounds:
            Iterable container of len 2 containing positive lower and upper
            bounds (in that order), or the string `"fixed"`.

    Raises:
        ValueError:
            Any nonpositive `bounds` string will produce an error.
    """

    def __init__(
        self,
        val: Union[str, float],
        bounds: Union[str, Tuple[float, float]] = "fixed",
    ):
        super(HomoscedasticNoise, self).__init__(val, bounds)
        if self.fixed() is False:
            if self._bounds[0] < 0.0 or self._bounds[1] < 0.0:
                raise ValueError(
                    f"Homoscedastic noise optimization bounds {self._bounds} "
                    f"are not strictly positive!"
                )



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

import MuyGPyS._src.math.torch as torch


def _homoscedastic_perturb(K: torch.ndarray, eps: float) -> torch.ndarray:
    _, nn_count, _ = K.shape
    return K + eps * torch.eye(nn_count)


def _heteroscedastic_perturb(
    K: torch.ndarray, eps: torch.ndarray
) -> torch.ndarray:
    ret = K.clone()
    batch_count, nn_count, _ = K.shape
    indices = (
        torch.repeat(torch.arange(batch_count), nn_count),
        torch.arange(nn_count).repeat(batch_count),
        torch.arange(nn_count).repeat(batch_count),
    )
    ret[indices] += eps.flatten()
    return ret



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
TensorHyperparameters

TensorHyperparameters specifications are expected to be provided in `Dict` form with
the key `"val"`. `"val"` is either an mm.ndarray value.
"""

from typing import Callable, Optional, Tuple, Type

import MuyGPyS._src.math as mm


class TensorHyperparameter:
    """
    A MuyGPs kernel or model Tensor Hyperparameter.

    TensorHyperparameters are defined solely by a value. Values must
    be scalar numeric arrays. Currently only used for heteroscedastic noise.

    Args:
        val:
            A mm.ndarray containing the value of the tensor hyperparameter
    """

    def __init__(
        self,
        val: mm.ndarray,
    ):
        """
        Initialize a tensor hyperparameter.
        """
        self._set_val(val)

    def _set(
        self,
        val: Optional[mm.ndarray] = None,
    ) -> None:
        """
        Reset hyperparameter value using keyword arguments.

        Args:
            val:
                A valid value.
        """
        if val is not None:
            self._set_val(val)

    def _set_val(self, val: mm.ndarray) -> None:
        """
        Set tensor hyperparameter value.

        Throws on out-of-range and other badness.

        Args:
            val:
                A valid mm.ndarray value.

        Raises:
            ValueError:
                A non-numeric, non-fixed, or string val
                will produce an error.
        """
        if isinstance(val, str):
            raise ValueError(
                "TensorHyperparameter class does not support strings."
            )
        if not isinstance(val, mm.ndarray):
            raise ValueError(
                f"Non-array tensor hyperparameter value {val} is not allowed."
            )
        if self.fixed() is False:
            raise ValueError(
                "TensorHyperparameters objects do not support optimization."
            )
        self._val = val

    def __call__(self) -> mm.ndarray:
        """
        Value accessor.

        Returns:
            The current value of the tensor hyperparameter.
        """
        return self._val

    def fixed(self) -> bool:
        """
        Report whether the parameter is fixed, and is to be ignored during
        optimization. Always returns True for tensor hyperparameters.

        Returns:
            `True`.
        """
        return True

    def get_bounds(self) -> Tuple[float, float]:
        raise NotImplementedError(
            "TensorHyperparameter does not support optimization bounds!"
        )

    def apply(self, fn: Callable, name: str) -> Callable:
        if self.fixed():

            def applied_fn(*args, **kwargs):
                kwargs.setdefault(name, self())
                return fn(*args, **kwargs)

            return applied_fn

        return fn

    def append_lists(self, name, names, params, bounds):
        pass


def _init_tensor_hyperparameter(
    val_def: mm.ndarray,
    type: Type = TensorHyperparameter,
    **kwargs,
) -> TensorHyperparameter:
    """
    Initialize a tensor hyperparameter given default value.

    Args:
        val:
            A valid value.
        kwargs:
            A hyperparameter dict including as subset of the key `val`.
    """
    val = kwargs.get("val", val_def)
    return type(val)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Tensor functions

Compute pairwise and crosswise difference tensors for the purposes of kernel
construction.

See the following example computing the pairwise and crosswise differences
between a batch of training data and their nearest neighbors.

Example:
    >>> from MuyGPyS.neighbors import NN_Wrapper
    >>> from MuyGPyS.optimize.batch import sample_batch
    >>> from MuyGPyS.gp.tensors import crosswise_tensor, pairwise_tensor
    >>> train_features = load_train_features()
    >>> nn_count = 10
    >>> nbrs_lookup = NN_Wrapper(
    ...         train_features,
    ...         nn_count,
    ...         nn_method="exact",
    ...         algorithm="ball_tree",
    ... )
    >>> train_count, _ = train_features.shape
    >>> batch_count = 50
    >>> batch_indices, batch_nn_indices = sample_batch(
    ...         nbrs_lookup, batch_count, train_count
    ... )
    >>> pairwise_diffs = pairwise_tensor(
    ...         train_features, batch_nn_inidices
    ... )
    >>> crosswise_diffs = crosswise_tensor(
    ...         train_features,
    ...         train_features,
    ...         batch_indices,
    ...         batch_nn_indices,
    ... )

See also the following example computing the crosswise differences between a
test dataset and their nearest neighors in the training data.

Example:
    >>> from MuyGPyS.neighbors import NN_Wrapper
    >>> from MuyGPyS.gp.tensors import crosswise_tensor, pairwise_tensor
    >>> train_features = load_train_features()
    >>> test_features = load_test_features()
    >>> nn_count = 10
    >>> nbrs_lookup = NN_Wrapper(
    ...         train_features, nn_count, nn_method="exact", algorithm="ball_tree"
    ... )
    >>> nn_indices, nn_diffs = nbrs_lookup.get_nns(test_features)
    >>> test_count, _ = test_features.shape
    >>> indices = np.arange(test_count)
    >>> nn_indices, _ = nbrs_lookup.get_nns(test_features)
    >>> pairwise_diffs = pairwise_tensor(
    ...         train_features, nn_inidices
    ... )
    >>> crosswise_diffs = crosswise_tensor(
    ...         test_features,
    ...         train_features,
    ...         indices,
    ...         nn_indices,
    ... )

The helper functions :func:`MuyGPyS.gp.tensors.make_predict_tensors`,
:func:`MuyGPyS.gp.tensors.make_fast_predict_tensors`, and
:func:`MuyGPyS.gp.tensors.make_train_tensors` wrap these difference tensors and
also return the nearest neighbors sets' training targets and (in the latter
case) the training targets of the training batch. These functions are convenient
as the difference and target tensors are usually needed together.
"""

from typing import Optional, Tuple

import MuyGPyS._src.math as mm
from MuyGPyS._src.gp.tensors import (
    _make_fast_predict_tensors,
    _make_predict_tensors,
    _make_train_tensors,
    _batch_features_tensor,
    _crosswise_tensor,
    _pairwise_tensor,
    _fast_nn_update,
    _make_heteroscedastic_tensor,
)


def make_heteroscedastic_tensor(
    measurement_noise: mm.ndarray,
    batch_nn_indices: mm.ndarray,
) -> mm.ndarray:
    """
    Create the heteroscedastic noise tensor for nonuniform noise values.

    Used to produce the noise tensor needed during batched training and
    prediction. Creates the `eps_tensor` tensor required by heteroscedastic
    MuyGPs models.

    Args:
        measurement_noise:
            A matrix of floats of shape `(batch_count,)` providing the noise
            corresponding to the response variable at each input value in the
            data.
        batch_nn_indices:
            A matrix of integers of shape `(batch_count, nn_count, nn_count)`
            listing the measurement noise for the nearest neighbors for all
            observations in the batch.

    Returns:
        A matrix of floats of shape `(batch_count, nn_count)` providing the
        noise corresponding to the nearest neighbor responses for all
        observations in the batch.
    """
    return _make_heteroscedastic_tensor(measurement_noise, batch_nn_indices)


def fast_nn_update(
    train_nn_indices: mm.ndarray,
) -> mm.ndarray:
    """
    Modify the nearest neighbor indices of the training data to include self.

    This function is only intended for use in concert with
    :func:~MuyGPyS.gp.tensors.make_fast_predict_tensors` and
    :func:`MuyGPyS.gp.muygps.MuyGPS.fast_coefficients`.

    Example:
        >>> train_nn_indices, _ = nbrs_lookup.get_nns(train_features)
        >>> train_nn_indices = fast_nn_update(train_nn_indices)
        >>> pairwise_diffs, nn_targets = make_fast_predict_tensors(
        ...     train_nn_indices,
        ...     train_features,
        ...     train_responses,
        ... )
        >>> K = muygps_fast.kernel(pairwise_diffs)
        >>> precomputed_coefficients_matrix = muygps_fast.fast_coefficients(
        ...     K, nn_targets
        ... )
        >>> # Late on, once test data is encountered
        >>> test_indices = np.arange(test_count)
        >>> test_nn_indices, _ = nbrs_lookup.get_nns(test_features)
        >>> closest_neighbor = test_nn_indices[:, 0]
        >>> closest_set = train_nn_indices[closest_neighbor, :]

    Args:
        train_nn_indices:
            A matrix of integers of shape `(train_count, nn_count)` listing the
            nearest neighbor indices for all observations in the batch.

    Returns:
        An integral matrix of shape `(train_count, nn_count)` that is similar
        to the input, but the most distant neighbor index is removed and the
        ``self'' index has been inserted.
    """
    return _fast_nn_update(train_nn_indices)


def make_fast_predict_tensors(
    batch_nn_indices: mm.ndarray,
    train_features: mm.ndarray,
    train_targets: mm.ndarray,
) -> Tuple[mm.ndarray, mm.ndarray]:
    """
    Create the difference and target tensors for fast posterior mean inference.

    Creates `pairwise_diffs` and `batch_nn_targets` tensors required by
    :func:`MuyGPyS.gp.muygps.MuyGPS.fast_posterior_mean`.

    Args:
        batch_nn_indices:
            A matrix of integers of shape `(batch_count, nn_count)` listing the
            nearest neighbor indices for all observations in the batch.
        train_features:
            The full floating point training data matrix of shape
            `(train_count, feature_count)`.
        train_targets:
            A matrix of shape `(train_count, response_count)` whose rows are
            vector-valued responses for each training element.

    Returns
    -------
    pairwise_diffs:
        A tensor of shape `(batch_count, nn_count, nn_count, feature_count)`
        containing the `(nn_count, nn_count, feature_count)`-shaped pairwise
        nearest neighbor difference tensors corresponding to each of the
        batch elements.
    batch_nn_targets:
        Tensor of floats of shape `(batch_count, nn_count, response_count)`
        containing the expected response for each nearest neighbor of each batch
        element.
    """
    return _make_fast_predict_tensors(
        batch_nn_indices,
        train_features,
        train_targets,
    )


def make_predict_tensors(
    batch_indices: mm.ndarray,
    batch_nn_indices: mm.ndarray,
    test_features: Optional[mm.ndarray],
    train_features: mm.ndarray,
    train_targets: mm.ndarray,
) -> Tuple[mm.ndarray, mm.ndarray, mm.ndarray]:
    """
    Create the difference and target tensors for prediction.

    Creates the `crosswise_diffs`, `pairwise_diffs` and `batch_nn_targets`
    tensors required by :func:`~MuyGPyS.gp.MuyGPS.posterior_mean` and
    :func:`~MuyGPyS.gp.MuyGPS.posterior_variance`.

    Args:
        batch_indices:
            A vector of integers of shape `(batch_count,)` identifying the
            training batch of observations to be approximated.
        batch_nn_indices:
            A matrix of integers of shape `(batch_count, nn_count)` listing the
            nearest neighbor indices for all observations in the batch.
        test_features:
            The full floating point testing data matrix of shape
            `(test_count, feature_count)`.
        train_features:
            The full floating point training data matrix of shape
            `(train_count, feature_count)`.
        train_targets:
            A matrix of shape `(train_count, feature_count)` whose rows are
            vector-valued responses for each training element.

    Returns
    -------
    crosswise_diffs:
        A tensor of shape `(batch_count, nn_count, feature_count)` whose last
        two dimensions list the difference between each feature of each batch
        element element and its nearest neighbors.
    pairwise_diffs:
        A tensor of shape `(batch_count, nn_count, nn_count, feature_count)`
        containing the `(nn_count, nn_count, feature_count)`-shaped pairwise
        nearest neighbor difference tensors corresponding to each of the
        batch elements.
    batch_nn_targets:
        Tensor of floats of shape `(batch_count, nn_count, response_count)`
        containing the expected response for each nearest neighbor of each batch
        element.
    """
    return _make_predict_tensors(
        batch_indices,
        batch_nn_indices,
        test_features,
        train_features,
        train_targets,
    )


def make_train_tensors(
    batch_indices: mm.ndarray,
    batch_nn_indices: mm.ndarray,
    train_features: mm.ndarray,
    train_targets: mm.ndarray,
) -> Tuple[mm.ndarray, mm.ndarray, mm.ndarray, mm.ndarray]:
    """
    Create the difference and target tensors needed for training.

    Similar to :func:`~MuyGPyS.gp.data.make_predict_tensors` but returns the
    additional `batch_targets` matrix, which is only defined for a batch of
    training data.

    Args:
        batch_indices:
            A vector of integers of shape `(batch_count,)` identifying the
            training batch of observations to be approximated.
        batch_nn_indices:
            A matrix of integers of shape `(batch_count, nn_count)` listing the
            nearest neighbor indices for all observations in the batch.
        train_features:
            The full floating point training data matrix of shape
            `(train_count, feature_count)`.
        train_targets:
            A matrix of shape `(train_count, feature_count)` whose rows are
            vector-valued responses for each training element.

    Returns
    -------
    crosswise_diffs:
        A tensor of shape `(batch_count, nn_count, feature_count)` whose last
        two dimensions list the difference between each feature of each batch
        element element and its nearest neighbors.
    pairwise_diffs:
        A tensor of shape `(batch_count, nn_count, nn_count, feature_count)`
        containing the `(nn_count, nn_count, feature_count)`-shaped pairwise
        nearest neighbor difference tensors corresponding to each of the batch
        elements.
    batch_targets:
        Matrix of floats of shape `(batch_count, response_count)` whose rows
        give the expected response for each batch element.
    batch_nn_targets:
        Tensor of floats of shape `(batch_count, nn_count, response_count)`
        containing the expected response for each nearest neighbor of each batch
        element.
    """
    return _make_train_tensors(
        batch_indices, batch_nn_indices, train_features, train_targets
    )


def batch_features_tensor(
    features: mm.ndarray,
    batch_indices: mm.ndarray,
) -> mm.ndarray:
    """
    Compute a tensor of feature vectors for each batch element.

    Args:
        features:
            The full floating point training or testing data matrix of shape
            `(train_count, feature_count)` or `(test_count, feature_count)`.
        batch_indices:
            A vector of integers of shape `(batch_count,)` identifying the
            training batch of observations to be approximated.

    Returns:
        A tensor of shape `(batch_count, feature_count)` containing
        the feature vectors for each batch element.
    """
    return _batch_features_tensor(features, batch_indices)


def crosswise_tensor(
    data: mm.ndarray,
    nn_data: mm.ndarray,
    data_indices: mm.ndarray,
    nn_indices: mm.ndarray,
) -> mm.ndarray:
    """
    Compute a matrix of differences between data and their nearest neighbors.

    Takes full datasets of records of interest `data` and neighbor candidates
    `nn_data` and produces the differences between each element of `data`
    indicated by `data_indices` and each of the nearest neighbors
    in `nn_data` as indicated by the corresponding rows of `nn_indices`. `data`
    and `nn_data` can refer to the same dataset.

    See the following example computing the crosswise differences between a
    batch of training data and their nearest neighbors.

    Args:
        data:
            The data matrix of shape `(data_count, feature_count)` containing
            batch elements.
        nn_data:
            The data matrix of shape `(candidate_count, feature_count)`
            containing the universe of candidate neighbors for the batch
            elements. Might be the same as `data`.
        indices:
            An integral vector of shape `(batch_count,)` containing the indices
            of the batch.
        nn_indices:
            An integral matrix of shape (batch_count, nn_count) listing the
            nearest neighbor indices for the batch of data points.

    Returns:
        A tensor of shape `(batch_count, nn_count, feature_count)` whose last
        two dimensions list the difference between each feature of each batch
        element element and its nearest neighbors.
    """
    return _crosswise_tensor(data, nn_data, data_indices, nn_indices)


def pairwise_tensor(
    data: mm.ndarray,
    nn_indices: mm.ndarray,
) -> mm.ndarray:
    """
    Compute a tensor of pairwise differences among sets of nearest neighbors.

    Takes a full dataset of records of interest `data` and produces the
    pairwise differences between the elements indicated by each row of
    `nn_indices`.

    Args:
        data:
            The data matrix of shape `(batch_count, feature_count)` containing
            batch elements.
        nn_indices:
            An integral matrix of shape (batch_count, nn_count) listing the
            nearest neighbor indices for the batch of data points.

    Returns:
        A tensor of shape `(batch_count, nn_count, nn_count, feature_count)`
        containing the `(nn_count, nn_count, feature_count)`-shaped pairwise
        nearest neighbor difference tensors corresponding to each of the
        batch elements.
    """
    return _pairwise_tensor(data, nn_indices)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from typing import Tuple

import MuyGPyS._src.math.torch as torch


def _make_heteroscedastic_tensor(
    measurement_noise: torch.ndarray,
    batch_nn_indices: torch.ndarray,
) -> torch.ndarray:
    eps_tensor = measurement_noise[batch_nn_indices]
    return eps_tensor


def _make_fast_predict_tensors(
    batch_nn_indices: torch.ndarray,
    train_features: torch.ndarray,
    train_targets: torch.ndarray,
) -> Tuple[torch.ndarray, torch.ndarray]:
    num_train, _ = train_features.shape
    batch_nn_indices_fast = torch.cat(
        (
            torch.unsqueeze(torch.arange(0, num_train), dim=1),
            batch_nn_indices[:, :-1],
        ),
        dim=1,
    )

    pairwise_dists_fast = _pairwise_tensor(
        train_features, batch_nn_indices_fast
    )
    batch_nn_targets_fast = train_targets[batch_nn_indices_fast]

    return pairwise_dists_fast, batch_nn_targets_fast


def _make_predict_tensors(
    batch_indices: torch.ndarray,
    batch_nn_indices: torch.ndarray,
    test_features: torch.ndarray,
    train_features: torch.ndarray,
    train_targets: torch.ndarray,
) -> Tuple[torch.ndarray, torch.ndarray, torch.ndarray]:
    if test_features is None:
        test_features = train_features
    crosswise_dists = _crosswise_tensor(
        test_features,
        train_features,
        batch_indices,
        batch_nn_indices,
    )
    pairwise_dists = _pairwise_tensor(train_features, batch_nn_indices)
    batch_nn_targets = train_targets[batch_nn_indices, :]
    return crosswise_dists, pairwise_dists, batch_nn_targets


def _make_train_tensors(
    batch_indices: torch.ndarray,
    batch_nn_indices: torch.ndarray,
    train_features: torch.ndarray,
    train_targets: torch.ndarray,
) -> Tuple[torch.ndarray, torch.ndarray, torch.ndarray, torch.ndarray]:
    crosswise_dists, pairwise_dists, batch_nn_targets = _make_predict_tensors(
        batch_indices,
        batch_nn_indices,
        train_features,
        train_features,
        train_targets,
    )
    batch_targets = train_targets[batch_indices, :]
    return crosswise_dists, pairwise_dists, batch_targets, batch_nn_targets


def _batch_features_tensor(
    features: torch.ndarray,
    batch_indices: torch.ndarray,
) -> torch.ndarray:
    return features[batch_indices, :]


def _crosswise_tensor(
    data: torch.ndarray,
    nn_data: torch.ndarray,
    data_indices: torch.ndarray,
    nn_indices: torch.ndarray,
) -> torch.ndarray:
    locations = data[data_indices]
    points = nn_data[nn_indices]
    return _crosswise_differences(locations, points)


def _pairwise_tensor(
    data: torch.ndarray,
    nn_indices: torch.ndarray,
) -> torch.ndarray:
    points = data[nn_indices]
    return _pairwise_differences(points)


def _crosswise_differences(
    locations: torch.ndarray, points: torch.ndarray
) -> torch.ndarray:
    return locations[:, None, :] - points


def _pairwise_differences(points: torch.ndarray) -> torch.ndarray:
    if len(points.shape) == 3:
        return points[:, :, None, :] - points[:, None, :, :]
    elif len(points.shape) == 2:
        return points[:, None, :] - points[None, :, :]
    else:
        raise ValueError(f"points shape {points.shape} is not supported.")


def _F2(diffs: torch.ndarray) -> torch.ndarray:
    return torch.sum(diffs**2, axis=-1)


def _l2(diffs: torch.ndarray) -> torch.ndarray:
    return torch.norm(diffs, dim=-1)


def _fast_nn_update(
    train_nn_indices: torch.ndarray,
) -> torch.ndarray:
    train_count, _ = train_nn_indices.shape
    new_nn_indices = torch.cat(
        (
            torch.unsqueeze(torch.arange(0, train_count), dim=1),
            train_nn_indices[:, :-1],
        ),
        dim=1,
    )
    return new_nn_indices



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from MuyGPyS import config


def fix_type(dtype):
    def typed_fn(fn):
        def fn_wrapper(*args, **kwargs):
            kwargs.setdefault("dtype", dtype)
            return fn(*args, **kwargs)

        return fn_wrapper

    return typed_fn


def fix_function_type(dtype, fn):
    return fix_type(dtype)(fn)


def fix_function_types(dtype, *fns):
    return tuple(fix_function_type(dtype, fn) for fn in fns)


def wrap_torch_signature(fn):
    def wrapped_fn(x, axis=0):
        return fn(x, dim=axis)

    return wrapped_fn


def wrap_torch_signatures(*fns):
    return tuple(wrap_torch_signature(fn) for fn in fns)


def set_type(t64, t32):
    if config.state.ftype == "64":
        return t64
    elif config.state.ftype == "32":
        return t32



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from typing import List, Tuple, Callable, Dict

import MuyGPyS._src.math as mm
from MuyGPyS._src.util import auto_str

from MuyGPyS.gp.hyperparameter import ScalarHyperparameter
from MuyGPyS.gp.hyperparameter.experimental import (
    HierarchicalNonstationaryHyperparameter,
)


@auto_str
class AnisotropicDistortion:
    """
    An isotropic distance model.

    IsotropicDistortion parameterizes a scaled elementwise distance function
    :math:`d(\\cdot, \\cdot)`, and is paramterized by a vector-valued
    :math:`\\mathbf{\\ell}>0` length scale hyperparameter.

    .. math::
         d_\\ell(\\mathbf{x}, \\mathbf{y}) =
         \\sum_{i=0}^d \\frac{d(\\mathbf{x}_i, \\mathbf{y}_i)}{\\ell_i}

    Args:
        metric:
            A callable metric function that takes a tensor of shape
            `(..., feature_count)` whose last dimension lists the elementwise
            differences between a pair of feature vectors and returns a tensor
            of shape `(...)`, having collapsed the last dimension into a
            scalar difference.
        length_scales:
            Keyword arguments `length_scale#`, mapping to scalar
            hyperparameters.
    """

    def __init__(
        self,
        metric: Callable,
        **length_scales,
    ):
        self._dist_fn = metric
        for i, key in enumerate(length_scales.keys()):
            if key != "length_scale" + str(i):
                raise ValueError(
                    "Anisotropic model expects one keyword argument for each "
                    "feature in the dataset labeled length_scale{i} for the "
                    "ith feature with indexing beginning at zero."
                )
        if not (
            all(
                isinstance(param, ScalarHyperparameter)
                for param in length_scales.values()
            )
            or all(
                isinstance(param, HierarchicalNonstationaryHyperparameter)
                for param in length_scales.values()
            )
        ):
            raise ValueError(
                "Anisotropic model expects all values for the length_scale{i} "
                "keyword arguments to be of the same type, either "
                "ScalarHyperparameter or HierarchicalNonstationaryHyperparameter."
            )
        self.length_scale = length_scales

    def __call__(
        self, diffs: mm.ndarray, batch_features=None, **length_scales
    ) -> mm.ndarray:
        """
        Apply anisotropic distortion to an elementwise difference tensor.

        This function is not intended to be invoked directly by a user. It is
        instead functionally incorporated into some
        :class:`MuyGPyS.gp.kernels.KernelFn` in its constructor.

        Args:
            diffs:
                A tensor of pairwise differences of shape
                `(..., feature_count)`.
            batch_features:
                A `(batch_count, feature_count)` matrix of features to be used
                with a hierarchical hyperparameter. `None` otherwise.
            length_scale:
                A floating point length scale, or a vector of `(knot_count,)`
                knot length scales.
        Returns:
            A crosswise distance matrix of shape `(data_count, nn_count)` or a
            pairwise distance tensor of shape
            `(data_count, nn_count, nn_count)` whose last two dimensions are
            pairwise distance matrices.
        """
        length_scale_array = self._get_length_scale_array(
            mm.array, diffs.shape, batch_features, **length_scales
        )
        return self._dist_fn(diffs / length_scale_array)

    @staticmethod
    def _get_length_scale_array(
        array_fn: Callable,
        target_shape: mm.ndarray,
        batch_features=None,
        **length_scales,
    ) -> mm.ndarray:
        # NOTE[MWP] THIS WILL NOT WORK WITH TORCH OPTIMIZATION.
        # We need to eliminate the implicit copy. Will need indirection.
        # We should make this whole workflow ifless.
        AnisotropicDistortion._lengths_agree(
            len(length_scales), target_shape[-1]
        )
        if callable(length_scales["length_scale0"]) is True:
            length_scales = {
                ls: length_scales[ls](batch_features) for ls in length_scales
            }
        # make sure each length_scale array is broadcastable when its shape is (batch_count,)
        shape = (1,) * (len(target_shape) - 2) + (-1,)
        return array_fn(
            [mm.reshape(value, shape) for value in length_scales.values()]
        ).T

    @staticmethod
    def _lengths_agree(found_length, target_length):
        if target_length != found_length and found_length != 1:
            raise ValueError(
                f"Number of lengthscale parameters ({found_length}) "
                f"must match number of features ({target_length}) or be 1 "
                "(Isotropic model)."
            )

    def get_opt_params(
        self,
    ) -> Tuple[List[str], List[float], List[Tuple[float, float]]]:
        """
        Report lists of unfixed hyperparameter names, values, and bounds.

        Returns
        -------
            names:
                A list of unfixed hyperparameter names.
            params:
                A list of unfixed hyperparameter values.
            bounds:
                A list of unfixed hyperparameter bound tuples.
        """
        names: List[str] = []
        params: List[float] = []
        bounds: List[Tuple[float, float]] = []
        for name, param in self.length_scale.items():
            param.append_lists(name, names, params, bounds)
        return names, params, bounds

    def get_opt_fn(self, fn) -> Callable:
        """
        Return a kernel function with fixed parameters set.

        This function is designed for use with
        :func:`MuyGPyS.optimize.chassis.optimize_from_tensors()` and assumes
        that optimization parameters will be passed as keyword arguments.

        Returns:
            A function implementing the kernel where all fixed parameters are
            set. The function expects keyword arguments corresponding to current
            hyperparameter values for unfixed parameters.
        """
        for name, param in self.length_scale.items():
            fn = param.apply(fn, name)
        opt_fn = fn
        return opt_fn

    def populate_length_scale(self, hyperparameters: Dict) -> None:
        """
        Populates the hyperparameter dictionary of a KernelFn object with
        `self.length_scales` of the AnisotropicDistortion object.

        Args:
            hyperparameters:
                A dict containing the hyperparameters of a KernelFn object.
        """
        for key, param in self.length_scale.items():
            hyperparameters[key] = param



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

"""
Noise modeling

Defines data structures and functors that handle noise priors for MuyGPs models.
"""
import MuyGPyS._src.math as mm

from MuyGPyS.gp.hyperparameter import TensorHyperparameter


class HeteroscedasticNoise(TensorHyperparameter):
    """
    A tensor :math:`\\eps` noise parameter.

    :math:`\\epse` is a heteroscedastic noise tensor used to build the "nugget"
    with the prior assumption that all observations are have a corresponding
    measurement noise.

    Args:
        val:
            An ndarray of shape `(batch_count, nn_count)`
            containing the heteroscedastic nugget matrix.
    Raises:
        ValueError:
            Any strictly negative entry in the array will produce an error.
    """

    def __init__(
        self,
        val: mm.ndarray,
    ):
        super(HeteroscedasticNoise, self).__init__(val)
        if mm.sum(self._val.flatten() < 0) > 0:
            raise ValueError(
                "Heteroscedastic noise values are not strictly non-negative!"
            )

    def fixed(self) -> bool:
        """
        Overloading fixed function to return True for heteroscedastic noise.

        Returns:
            `True` - we do not allowed
        """
        return True



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from setuptools import setup


TEST_REQUIRES = [
    "absl-py>=0.13.0",
]

DEV_REQUIRES = [
    "black>=21.1.0",
    "build>=0.7.0",
    "mypy>=0.910",
    "twine>=3.7.1",
    "h5py>=3.7.0",
]

DOCS_REQUIRES = [
    "sphinx==6.2.1",
    "sphinx-rtd-theme==1.2.2",
    "sphinx-autodoc-typehints==1.22",
    "matplotlib>=3.2.1",
    "nbsphinx==0.9.2",
    "pandas==1.5.2",
    "pandoc==2.3.0",
    "pandocfilters==1.5.0",
    "ipython==7.30.1",
    "ipykernel==6.6.0",
    "torchvision>=0.14.1",
]

HNSWLIB_REQUIRES = [
    "pybind11>=2.5.0",
    "hnswlib>=0.6.0",
]

JAX_REQUIRES = [
    "tensorflow-probability[jax]>=0.16.0",
]

JAX_CPU_REQUIRES = [
    "jax[cpu]>=0.2.26",
]

MPI_REQUIRES = [
    "mpi4py==3.1.3",
]

TORCH_REQUIRES = [
    "torch>=1.13.0",
]

DOCS_REQUIRES += TORCH_REQUIRES

setup(
    extras_require={
        "dev": DEV_REQUIRES + TEST_REQUIRES + DOCS_REQUIRES,
        "docs": DOCS_REQUIRES,
        "tests": TEST_REQUIRES,
        "hnswlib": HNSWLIB_REQUIRES,
        "jax_cpu": JAX_CPU_REQUIRES + JAX_REQUIRES,
        "mpi": MPI_REQUIRES,
        "torch": TORCH_REQUIRES,
    },
)



# Copyright 2021-2023 Lawrence Livermore National Security, LLC and other
# MuyGPyS Project Developers. See the top-level COPYRIGHT file for details.
#
# SPDX-License-Identifier: MIT

from MuyGPyS._src.util import _collect_implementation

(
    _mse_fn,
    _cross_entropy_fn,
    _lool_fn,
    _lool_fn_unscaled,
    _pseudo_huber_fn,
    _looph_fn,
) = _collect_implementation(
    "MuyGPyS._src.optimize.loss",
    "_mse_fn",
    "_cross_entropy_fn",
    "_lool_fn",
    "_lool_fn_unscaled",
    "_pseudo_huber_fn",
    "_looph_fn",
)
